//***********************************************************************

//
//
//                              GENSAMPLE.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : sampling functions used by FISPRO, part of library fispro

//**********************************************************************



//*********************************************************************
//
//
//                Common functions for FIS classes
//
//
//*********************************************************************


#include "common.h"
#include "gensample.h"
#include <math.h>  
#include <stdexcept>
#include <stdlib.h>
#include <cstdlib>
#include <ctime>

//***************************  gensample.cpp  **********************
int  intcompare(const void *p1, const void *p2)
     {
         int i = *((int *)p1);
         int j = *((int *)p2);

         if (i > j)
             return (1);
         if (i < j)
             return (-1);
         return (0);
     }
//*****************************************************************************
void genNSamples(long int seed,double percV,int classif,double eps, int outCol, char * datafile, int nfiles,int flagDisplay)
{
  //! generates nfiles pairs of sample validation and sample test files
  //! calls Alloc1DDoubleWorkingArray,Alloc1DIntWorkingArray, Alloc2DDoubleWorkingArray
  //! calls ReadSampleFile to read data array
  //! if classif flag, calls SortUniq to find classes in data array outCol
  //! in that case eps is the precision to discriminate classes
  //! calls writeVTSample (classif case) or genVTSample to write samples
  //! the only difference between the two is that genVTSample opens and closes the files
  //! while writeVTSample appends to them
  //! the seed argument allows to reproduce a random sequence
  //! except if seed=0, new random sequence based on machine clock
  
  
  int i,nr,ncol,nrow,nf,sizeV,nuniq,res,sizecumV,sizesubV,clas,ind,j;
  int * occur=NULL, *groupIndex=NULL;
  double percsubV;
  double ** Examples=NULL, ** subExamples=NULL, *uniq=NULL;
  double * col=NULL;
  char * fileVname=NULL, *fileTname=NULL;
  FILE * fileV, * fileT;

  // check arguments
  if (percV<=0. || percV>0.9)
      percV=0.9;
  if(seed==0) 
      seed=time(NULL);//machine clock
  if (nfiles<=0)
    throw std::runtime_error("~number~of~samples~must~be~>0");
  if (nfiles>99)
    throw std::runtime_error("~number~of~samples~limited~to~99");
  if (flagDisplay) printf("\nseed=%ld ",seed);
  setseed(seed);
    
  // init arrays
  Examples=ReadSampleFile(datafile,ncol,nrow);
  sizeV=( int )floor (nrow*percV);
  if (flagDisplay) printf("data sample size=%d rows, generation of %d sample(s), validation sample size=%d rows\n",nrow,nfiles,sizeV);
  // init classif. case
  // find nuniq and init classif arrays:  uniq, occur
  if (classif)
	{
	  // find classes in data column (outCol)
	  if (flagDisplay) printf("\nClassification sample\n");
	  if(outCol==-1) 
	    outCol=ncol-1;
	  else
	    if (outCol<0 || outCol>ncol)
	      throw std::runtime_error("~incorrect~column~number");
	  
	  col=Alloc1DDoubleWorkingArray(nrow); //alloc and init 0.0
	  groupIndex=Alloc1DIntWorkingArray(nrow);//alloc and init 0
	  
	  for(nr = 0; nr <nrow ; nr ++)   
	    {
	      //chosen column
	      col[nr]=Examples[nr][outCol];  
	    }
	  
	  res=SortUniq(col,nrow,&uniq,&occur,nuniq,eps); // col is sorted out
	  if (res==-1) throw std::runtime_error("~NA~values~in~data~set");
	  //unique values are in uniq, number of occurences are in nuniq
	  for (clas=0;clas<nuniq;clas++)
	    {
	      if(flagDisplay) printf("\n class %d, value=%f,  occurrence #=%d",clas,uniq[clas],occur[clas]);
	      if(flagDisplay) printf("\n");
	      percsubV=percV*occur[clas]/nrow;
	      sizesubV=( int )floor (percsubV*nrow);
	      if (sizesubV==0) 
		{
		  sprintf(ErrorMsg,"Not~enough~examples~in~data~set~class#%d~to~draw~sample",clas);
		  throw std::runtime_error(ErrorMsg);
		}
	    }
	   // check that we have enough examples in data set
	  if (sizeV<nuniq) throw std::runtime_error("Not~enough~examples~in~data~set~to~draw~sample");

	} // classif. case init

  //file loop - 2 sample files are written at each iteration
  // in classif. case, subExamples arrays are rebuilt at each iteration (to save memory)
  for (nf=0;nf<nfiles;nf++)
    {
      // generate file names for valid. and test samples
      fileVname=new char[strlen(datafile) +22];
      if (fileVname==NULL) throw std::runtime_error("~memory~allocation~error");
      sprintf(fileVname,"%s.lrn.sample.%d",datafile,nf);
  
      fileTname=new char[strlen(datafile) +22];
      if (fileTname==NULL) throw std::runtime_error("~memory~allocation~error");
      sprintf(fileTname,"%s.tst.sample.%d",datafile,nf);
      if (flagDisplay)
	{
	  printf("\n----Sample #%d----------------------------------------",nf);
	  printf("\nfileVname=%s\tfileTname=%s",fileVname,fileTname);
	}

      if (classif)
	{
	  
	  fileV=openFileW(fileVname); 
	  fileT=openFileW(fileTname);

	  
	  //now build a subarray for each class and draw a random sample from it.
	  // append successive samples to files
	  sizecumV=0;
	  for (clas=0;clas<nuniq;clas++)
	    {
	      subExamples=Alloc2DDoubleWorkingArray(occur[clas],ncol);
	      //groupIndex contains for each row the class index
	      ind=0;
	      i=0;
	      
	      while(i<nrow && ind<occur[clas]) // double protection
		{
		  if (fabs(Examples[i][outCol]-uniq[clas])<eps)
		    {
		      groupIndex[i]=clas;
		      for (j=0;j<ncol;j++)
			subExamples[ind][j]=Examples[i][j];
		      ind++;
		    }
		  i++;
		}

	      if (flagDisplay) 
		{
		  
		  printf("\nSubsample for class %d, size=%d\n",clas,occur[clas]);
		  for(i=0;i<occur[clas];i++)
		    {
		      for (j=0;j<ncol;j++)
			printf("%f  ",  subExamples[i][j]);
		      printf("\n");
		    }
		}
	      
	      percsubV=percV*occur[clas]/nrow;
	      sizesubV=( int )floor (percsubV*nrow);
	      if(clas==(nuniq-1)) sizesubV=sizeV-sizecumV;

	      if (flagDisplay)
		{
		  printf("\nclass %d",clas);
		  printf("\n------------------------------------------------------------");
		}
	      

	      sizecumV+=sizesubV;
	      writeVTSample(subExamples,occur[clas],ncol,sizesubV,fileV,fileT,flagDisplay);

	      // delete subExample, uniq and occur arrays before next iteration
	      del2DArray(subExamples,occur[clas]);
	    }
	    
	  fclose(fileV);
	  fclose(fileT);
	} // end classif. case

      else
	  genVTSample(Examples,nrow,ncol,sizeV,fileVname,fileTname,flagDisplay);

      if (fileVname !=NULL) delete [] fileVname;
      if (fileTname !=NULL) delete [] fileTname;
  

    } // end file loop
  if (groupIndex!=NULL) delete [] groupIndex;
  if (col!=NULL) delete [] col;
  if (occur !=NULL) delete [] occur;
  if (uniq !=NULL) delete [] uniq;
  del2DArray(Examples,nrow);
  
} //end genNSamples function

//*****************************************************************************
void genVTSample(double ** Examples, int nrow, int ncol, int sizeV, char * fileVname,char * fileTname,int flagDisplay)
  //! opens files 
  //! and calls writeVTSample which generates 1 valid. sample (sizeV) and 1 test. sample (remaining items) from Examples array
  //! and writes them to the files
{

  // check files
 
  
  FILE * fileV=NULL, *fileT=NULL;

  fileV=openFileW(fileVname);
  fileT=openFileW(fileTname);
 
  // call sampling function
  // mimimum size=1, maximum size=nrow
  if (sizeV<=0) sizeV=1;
  if (sizeV>nrow)sizeV=nrow;
 
  writeVTSample(Examples,nrow,ncol,sizeV,fileV,fileT,flagDisplay);
  
fclose(fileV);
fclose(fileT);
}

//*****************************************************************************
void writeVTSample(double ** Examples, int nrow, int ncol, int sizeV, FILE * fileV, FILE * fileT,int flagDisplay)
{
  //! files are already opened - appends values to them
  //! values are drawn randomly from the Examples array by finding a series of row indexes
  //! calls FisRand 
  //! sizeV is the number of lines in the validation sample
  //! the rest will be the test sample
  
  
  double randVal,alea;
  int i,j,icol,nbelems,irand,uniqFlag,iter;
  int maxiter;
  int *irandVal=NULL;

  // check arguments
  if (nrow<=0 || ncol<=0) throw std::runtime_error("~incorrect~file~size");
  if(sizeV<0) throw std::runtime_error("~negative~sample~size");
  if(sizeV>nrow) throw std::runtime_error("~data~size~is~too~small~for~sample");
  if(Examples==NULL) throw std::runtime_error("~invalid~data~array");

  irandVal=Alloc1DIntWorkingArray(sizeV);
  
  // init algo
  alea=FisRand();
  randVal = floor (alea*nrow);
  irandVal[0]=(int) randVal;
  nbelems=1;
  iter=0;
  maxiter=sizeV*sizeV; // protection against infinite loop

  while ((nbelems<sizeV) && (iter<maxiter))
    { 
      // draw index
      iter++;
      alea=FisRand();//random double number between 0 and 1
      randVal = floor (alea*nrow);
      irand=(int) randVal;//transform into integer=row index
      uniqFlag=true;
      // check unicity of row index
      i=0;
      while((i<nbelems) && uniqFlag)
	{
	  if (irand==irandVal[i])
	    uniqFlag=false;
	  i++;
	}
      if (uniqFlag) 
	{
	  nbelems++;
	  irandVal[nbelems-1]=irand;
	}
    }
  //sort row indexes for learning file item selection
  qsort((void *)irandVal,sizeV , sizeof (int), intcompare);
  		  
  if (flagDisplay) 
    {	
      printf("\ngenerating learning  sample length=%d\n",sizeV);
      printf("indexes:\n");
      for (i=0;i<sizeV;i++)
	printf("%d ",irandVal[i]);
      printf("\n");
      printf("generating test sample length=%d",nrow-sizeV);
      printf("\n");
    }
  // read data array - write elements whose row indexes were selected in validation file
  j=0;
  for (i=0;i<nrow;i++)
    {
      if ((j<sizeV) && (i == irandVal[j]))
	{
	  //valid file 
	  for (icol=0;icol<ncol;icol++) 
	    {
	      fprintf(fileV,"%f",Examples[i][icol]); 
	      if (icol <(ncol-1)) fprintf(fileV," ,");
	    }
	  fprintf(fileV,"\n");
	  j++;
	}
      // write other elements in test file
      else
	{
	  //test file
	  if (flagDisplay) printf("%d ",i);
	  for (icol=0;icol<ncol;icol++) 
	    {
	      fprintf(fileT,"%f",Examples[i][icol]); 
	      if (icol <(ncol-1)) fprintf(fileT," ,");
	    }
	  fprintf(fileT,"\n");
	}
    }
  if (flagDisplay) printf("\n");

  // delete arrays
  if (irandVal!=NULL) delete [] irandVal;
}
//*****************************************************************************
void genSplit(long int seed, int truncate, int classif,double eps, int outCol,char * datafile, int nfiles,int flagDisplay)
{
  //! generates nfiles subsample  files
  //! calls Alloc1DDoubleWorkingArray,Alloc1DIntWorkingArray, Alloc2DDoubleWorkingArray
  //! calls ReadSampleFile to read data array
  //! calls writeVTSample (classif case) or genVTSample to write samples
  //! the only difference between the two is that genVTSample opens and closes the files
  //! while writeVTSample appends to them
  //! the seed argument allows to reproduce a random sequence
  //! except if seed=0, new random sequence based on machine clock
  
  
  int ncol,nrow,nf,sizesub,sizeV,sizecumV,nuniq,res,clas,nr,nrowInit, maxcl;
  int * occur=NULL,* occurInit=NULL, *sizesubV=NULL, *sizecumF=NULL ;
  double percsubV,percV;
  double ** Examples=NULL,  *** subExamples=NULL, *uniq=NULL;
  double * col=NULL;
  char * fileVname=NULL;
 
  
  // check arguments
  if(seed==0) 
      seed=time(NULL);//machine clock
  if (nfiles<=0)
    throw std::runtime_error("~number~of~samples~must~be~>0");
  if (nfiles>99)
    throw std::runtime_error("~number~of~samples~limited~to~99");
  if (flagDisplay) printf("\nseed=%ld ",seed);
  setseed(seed);
  percV=1.0/nfiles;  
  // init arrays
  Examples=ReadSampleFile(datafile,ncol,nrow);
  nrowInit=nrow;
  sizeV=( int )floor ((double)nrow/nfiles);
  if (flagDisplay) printf("data sample size=%d rows, generation of %d sample(s), sample size=%d rows\n",nrow,nfiles,sizeV);
  if (sizeV<5) 
    throw std::runtime_error("~sample~size~must~be~>=5~rows");
  //file loop -  sample files are written at each iteration
  // no resampling
  // 
  
// init classif. case
  // find nuniq and fill classif arrays:  uniq, occur
  if (classif)
	{
	  // find classes in data column (outCol)
	  if (flagDisplay) printf("\nClassification sample\n");
	  if(outCol==-1) 
	    outCol=ncol-1;
	  else
	    if (outCol<0 || outCol>ncol)
	      throw std::runtime_error("~incorrect~column~number");
	  
	  col=Alloc1DDoubleWorkingArray(nrow); //alloc and init 0.0
	  
	  for(nr = 0; nr <nrow ; nr ++)   
	    {
	      //chosen column
	      col[nr]=Examples[nr][outCol];  
	    }
	  
	  res=SortUniq(col,nrow,&uniq,&occur,nuniq,eps); // col is sorted out
	  if (res==-1) throw std::runtime_error("~NA~values~in~data~set");
	  //unique values are in uniq, number of occurences are in nuniq
	  maxcl=0;
	  for (clas=0;clas<nuniq;clas++)
	    {
	      if(occur[clas]>maxcl) maxcl=occur[clas];
	      if (flagDisplay) printf("\n class %d, value=%f,  occurrence #=%d",clas,uniq[clas],occur[clas]);
	      if (flagDisplay) printf("\n");
	      percsubV=percV*occur[clas]/nrow;
	      sizesub=( int )floor (percsubV*nrow);
	      if (sizesub==0) 
		{
		  sprintf(ErrorMsg,"Not~enough~examples~in~data~set~class#%d~to~draw~sample",clas);
		  throw std::runtime_error(ErrorMsg);
		}
	    }
	  if (flagDisplay) printf("maximum=%d\n",maxcl);
	   // check that we have enough examples in data set
	  if (sizeV<nuniq) throw std::runtime_error("Not~enough~examples~in~data~set~to~draw~sample");
	  sizecumF=Alloc1DIntWorkingArray(nuniq);
	  sizesubV=Alloc1DIntWorkingArray(nuniq);
	  // generate subsample 3d arrays for each class: class x nrow x ncol
	  // reorder classes so that the last one is the most numerous one
	  // shift class with maximum count to last position
	  // keep initial values of occur for each class
	  genSubSample(subExamples, nrow,  ncol,  outCol, eps, Examples,  uniq,  occur,  nuniq, maxcl,flagDisplay);
	  // save original count for each class
	  occurInit=Alloc1DIntWorkingArray(nuniq);
	  for (clas=0;clas<nuniq;clas++)
	    occurInit[clas]=occur[clas];
	 
	} // end classif. case init
  

  for (nf=0;nf<nfiles;nf++)
    {
      // generate file names for all samples
      fileVname=new char[strlen(datafile) +22];
      if (fileVname==NULL) throw std::runtime_error("~memory~allocation~error");
      sprintf(fileVname,"%s.sample.%d",datafile,nf);
  
      if (flagDisplay)
	{
	  printf("\n----Sample #%d----------------------------------------",nf);
	  printf("\nfileVname=%s\t",fileVname);
	}
      
      // generate sample and write it to file, update Examples array and nrow number
      if (classif)
	{
	  percV=1./(nfiles-nf);
	  FILE * fileV=NULL;
	  fileV=openFileW(fileVname);
	 
	  
	  sizecumV=0;
	  for (clas=0;clas<nuniq;clas++)
	    {
	      sizesubV[clas]=( int )floor (percV*occur[clas]);
	      //fill up each file with last class remaining elements
	      if(clas==(nuniq-1)) sizesubV[clas]=sizeV-sizecumV;
	      // last file = all remaining elements except if truncate=1
	      if((truncate==0) && (nf==(nfiles-1))) sizesubV[clas]=occurInit[clas]-sizecumF[clas];
	      // next call will update occur[clas] and subExamples[clas] array
	      
	      if ((sizesubV[clas] <0) || (sizesubV[clas]>(occurInit[clas]-sizecumF[clas]))) 
		{
		  printf("\n---%d files generated",nf);
		  throw std::runtime_error("~stopping~sampling~not~enough~items~");
		}
	      if (flagDisplay)
		{
		  printf("\nClass # %d \t total class items %d \t cumulated class size %d \t required size %d \t cum. sample size %d",clas,occurInit[clas],sizecumF[clas],sizesubV[clas], sizecumV);
		  
		}
	      writeVTSampleSplit(subExamples[clas],occur[clas],ncol,sizesubV[clas],fileV,flagDisplay);
	      // subExamples[clas] and occur[clas] have been updated by last call
	      sizecumV+=sizesubV[clas];
	      sizecumF[clas]+=sizesubV[clas];
	      	      
	    }
	   if (flagDisplay) printf("\n gen. sample size %d",sizecumV);
	  fclose(fileV);
	}
      else
	{
	  // last file = write all remaining elements except if truncate argument=1
	  if (flagDisplay) printf("\nnrow=%d",nrow);
	  if((truncate==0) && (nf==(nfiles-1))) sizeV=nrow;
	  genVTSampleSplit(Examples,nrow,ncol,sizeV,fileVname,flagDisplay);//updates nrow and Examples array
	}

      if (fileVname !=NULL) delete [] fileVname;
      

    } // end file loop
  
  if (col!=NULL) delete [] col;
  if (occur !=NULL) delete [] occur;
  if (uniq !=NULL) delete [] uniq;
  if ( sizecumF !=NULL) delete [] sizecumF;
  if ( sizesubV !=NULL) delete [] sizesubV;
  if ( occurInit!=NULL) delete [] occurInit;
  if (Examples!=NULL) del2DArray(Examples,nrow);
  if (subExamples!=NULL) del3DArray(subExamples,nuniq,nrowInit);
  
} //end genSplit function

//*****************************************************************************
void genVTSampleSplit(double ** Examples, int & nrow, int ncol, int sizeV, char * fileVname,int flagDisplay)
  //! opens files 
  //! and calls writeVTSampleSplit which generates 1 valid. sample (sizeV)  from Examples array
  //! and writes it to the file
{

  // check files
 
  
  FILE * fileV=NULL;

  fileV=openFileW(fileVname);
 
  // call sampling function
  // mimimum size=1, maximum size=nrow
  if (sizeV<=0) sizeV=1;
  if (sizeV>nrow)sizeV=nrow;
 
  writeVTSampleSplit(Examples,nrow,ncol,sizeV,fileV,flagDisplay);
  
fclose(fileV);
}

//*****************************************************************************
void writeVTSampleSplit(double ** Examples, int &nrow, int ncol, int sizeV, FILE * fileV, int flagDisplay)
{
  //! files are already opened - appends values to them
  //! values are drawn randomly from the Examples array by finding a series of row indexes
  //! calls FisRand 
  //! sizeV is the number of lines in the validation sample
  //! the rest will be for the next sample
  //! updates the nrow argument
  
  
  double randVal,alea;
  int i,j,sizeKeep,icol,nbelems,irand,uniqFlag,iter;
  int maxiter;
  int *irandVal=NULL;
  int *irem=NULL;

  // check arguments
  if (nrow<=0 || ncol<=0) throw std::runtime_error("~incorrect~file~size");
  if(sizeV<0) throw std::runtime_error("~negative~sample~size");
  //if(sizeV>nrow) throw std::runtime_error("~data~size~is~too~small~for~sample");
  if(Examples==NULL) throw std::runtime_error("~invalid~data~array");

  irandVal=Alloc1DIntWorkingArray(sizeV);
  if (nrow>sizeV)
    irem=Alloc1DIntWorkingArray(nrow-sizeV);
     

  // init algo
  alea=FisRand();
  randVal = floor (alea*nrow);
  irandVal[0]=(int) randVal;
  nbelems=1;
  iter=0;
  maxiter=sizeV*sizeV; // protection against infinite loop

  while ((nbelems<sizeV) && (iter<maxiter))
    { 
      // draw index
      iter++;
      alea=FisRand();//random double number between 0 and 1
      randVal = floor (alea*nrow);
      irand=(int) randVal;//transform into integer=row index
      uniqFlag=true;
      // check unicity of row index
      i=0;
      while((i<nbelems) && uniqFlag)
	{
	  if (irand==irandVal[i])
	    uniqFlag=false;
	  i++;
	}
      if (uniqFlag) 
	{
	  nbelems++;
	  irandVal[nbelems-1]=irand;
	}
    }
  //sort row indexes for file row selection
  qsort((void *)irandVal,sizeV , sizeof (int), intcompare);
  if (flagDisplay) printf("\ngenerating cross validation  sample length=%d\n",sizeV);
  if (flagDisplay) 
    {	
      printf("indexes:\n");
      for (i=0;i<sizeV;i++)
	printf("%d ",irandVal[i]);
      printf("\n");
    }
  if (flagDisplay)
    {
      printf("remaining sample length=%d",nrow-sizeV);
      printf("\nrows to be kept for next sampling:\n");
    }
  // read data array - write elements whose row indexes were selected in validation file
  j=0;
  i=0;
  sizeKeep=0;
  while(i<nrow)
    {
      if ((j<sizeV) && (i == irandVal[j]))
	{
	  //sample file 
	  for (icol=0;icol<ncol;icol++) 
	    {
	      fprintf(fileV,"%f",Examples[i][icol]); 
	      if (icol <(ncol-1)) fprintf(fileV," ,");
	    }
	  fprintf(fileV,"\n");
	  j++;
	}
      // elements to be kept 
      else
	{
	 
	  if (flagDisplay) printf("%d ",i);
	  //
	  if (irem !=NULL)
	    {
	      irem[sizeKeep]=i;
	      sizeKeep++;
	    }
	}
      i++;
    }
    
  if (flagDisplay) printf("\nnumber of remaining rows=%d\n",sizeKeep);

  
  // irem is already sorted
  // shift irem defined rows to beginning of Examples array 
  if (irem !=NULL)
    {
      for (i=0;i<sizeKeep;i++)
	for (j=0;j<ncol;j++)
	  Examples[i][j]=Examples[irem[i]][j];
    }
  //update nrow for next sample
  nrow=sizeKeep;
  // delete temporary arrays
  if (irandVal!=NULL) delete [] irandVal;
  if (irem!=NULL) delete [] irem;
  
}

//*****************************************************************************

FILE * openFileW(char * fileName)
{
  FILE * fileV;// open file for writing
  fileV=fopen(fileName,"wt");
  if (fileV==NULL)
     {
       sprintf( ErrorMsg, "~CannotOpenFile:%s~", fileName );
       throw std::runtime_error( ErrorMsg );
     }
  return(fileV);
}

//*****************************************************************************
void del2DArray(double ** x, int nrow)
{
  //
  int i;
  for (i=0;i<nrow;i++)
    if (x[i] !=NULL) delete [] x[i];
  if (x!=NULL) delete [] x;

}
//*****************************************************************************
void del3DArray(double *** x, int i1,int i2)
{
  //
  int j,i;
  for (i=0;i<i1;i++)
    {
      for (j=0;j<i2;j++)
	if (x[i][j] !=NULL) delete [] x[i][j];
      if (x[i] !=NULL) delete [] x[i];
    }
  if (x!=NULL) delete [] x;

}
//*****************************************************************************
double *** Alloc3DDoubleWorkingArray(int i1, int i2, int i3)
{
  //allocates 3d double array and initialize all elements to 0.
  // throws exception if memory allocation error
  int i,j,k;
  double *** tab3=NULL;

  if ((i1<1) || (i2 <1) || (i3 <1)) throw std::runtime_error("~required allocation~<0");
  tab3=new double**[i1];
  if (tab3==NULL) throw std::runtime_error("~memory~allocation~error");
  for (i=0;i<i1;i++)
    {
      tab3[i]=NULL;
      if (i2 >0) tab3[i]=new double*[i2];
      if (tab3[i]==NULL) throw std::runtime_error("~memory~allocation~error");
      
      for (j=0;j<i2;j++)
	{
	  tab3[i][j]=NULL;
	  if (i3 >0) tab3[i][j]=new double[i3];
	  if (tab3[i][j]==NULL) throw std::runtime_error("~memory~allocation~error");
	  for (k=0;k<i3;k++)
	    {
	      tab3[i][j][k]=0.; 
	    }
	}
    }
  return(tab3);
        
}
//*****************************************************************************
double ** Alloc2DDoubleWorkingArray(int nrow, int ncol)
{
  //allocates 2d double array and initialize all elements to 0.
  // throws exception if memory allocation error
  int k,kk;
  double ** mat=NULL;

  if ((nrow<1) || (ncol <1)) throw std::runtime_error("~required~allocation~<0");
  mat=new double*[nrow];
  if (mat==NULL) throw std::runtime_error("~memory~allocation~error");
  for (k=0;k<nrow;k++)
    {
      mat[k]=NULL;
      if (ncol >0) mat[k]=new double[ncol];
      if (mat[k]==NULL) throw std::runtime_error("~memory~allocation~error");
      
      for (kk=0;kk<ncol;kk++)
	{
	  mat[k][kk]=0.; 
	}
    }
  return(mat);
        
}
double * Alloc1DDoubleWorkingArray(int n)
{
  //*************************************************
  //allocates 1d double array and initialize all elements to 0.
  int k;
  double *vec;
  if (n<1) throw std::runtime_error("~required~allocation~<0");
  vec=new double[n];
  if (vec==NULL) throw std::runtime_error("~memory~allocation~error");
  for (k=0;k<n;k++)
    vec[k]=0.;
  return vec;
}
int * Alloc1DIntWorkingArray(int n)
{
  //*************************************************
  //allocates 1d int array and initialize all elements to 0.
  int k;
  int *vec;

  if (n<1) throw std::runtime_error("~required~allocation~<0");
  vec=new int[n];
  if (vec==NULL) throw std::runtime_error("~memory~allocation~error");
  for (k=0;k<n;k++)
    vec[k]=0;
  return vec;
}

void genSubSample(double *** & subExamples, int nrow, int ncol, int outCol, double eps, double ** Examples, double * uniq, int * occur, int nuniq, int maxcl, int flagDisplay)
{
  //*************************************************
  // fills subarrays for each class 
  // shift class with maximum count (maxcl) to last position
  int clas,i,j,ind,indmaxcl,oclast;
  double ** tmp=NULL;

  //now build a subarray for each class 
  subExamples=Alloc3DDoubleWorkingArray(nuniq,nrow,ncol);
  indmaxcl=0;
  oclast=0;
  for (clas=0;clas<nuniq;clas++)
    {
      
      ind=0;
      i=0;
      if(occur[clas]==maxcl) indmaxcl=clas;
	      
      while(i<nrow && ind<occur[clas]) // double protection
	{
	  if (fabs(Examples[i][outCol]-uniq[clas])<eps)
	    {
	      for (j=0;j<ncol;j++)
		subExamples[clas][ind][j]=Examples[i][j];
	      ind++;
	    }
	  i++;
	}
    }//end for

  // mow exchange last class with maxcl class
  tmp=Alloc2DDoubleWorkingArray(nrow,ncol);
  
  if (indmaxcl != (nuniq-1))
    {
      if (flagDisplay) printf("\nExchanging class %d with last class\n",indmaxcl);
      oclast=occur[nuniq-1];
      occur[indmaxcl]=oclast;
      occur[nuniq-1]=maxcl;
      for (i=0;i<nrow;i++)
	for (j=0;j<ncol;j++)
	  tmp[i][j]=subExamples[indmaxcl][i][j];
      for (i=0;i<nrow;i++)
	for (j=0;j<ncol;j++)
	  subExamples[indmaxcl][i][j]=subExamples[nuniq-1][i][j];
      for (i=0;i<nrow;i++)
	for (j=0;j<ncol;j++)
	  subExamples[nuniq-1][i][j]=tmp[i][j];
    }

if (flagDisplay) 
{
  for (clas=0;clas<nuniq;clas++)
    {	  
      printf("\nSubsample for class %d, size=%d\n",clas,occur[clas]);
      for(i=0;i<occur[clas];i++)
	{
	  for (j=0;j<ncol;j++)
	    printf("%f  ",  subExamples[clas][i][j]);
	  printf("\n");
	}
    }
  
}
 if (tmp !=NULL) del2DArray(tmp,nrow);
}
