//*********************************************************************
//
//
//                              PERF.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// Object : Call performance function
// Environnement : Uses library fispro

//*********************************************************************

#include "fis.h"

//! This program needs 2 arguments:  the fis configuration file name and 
//! the data file name. Options: -a for intermediate display ; 
//! -s activity threshold for an item not to be blank (default: 0.2).
int main(int argc, char ** argv)
//******************************
{
  
  if(argc < 3)
    {
      printf("\nThe program %s needs 2 arguments : ", argv[0]);
      printf("\n\t* the fis configuration file name;");   
      printf("\n\t* the data file name.") ;
      printf("\n\t options");
      printf("\n\t -un Normalized distance (default normalized)");
      printf("\n\t -m Minkowski power (default 2)");
      printf("\n\t -c '0 1' string Euclidean/fuzzy (default fuzzy distance)");
      printf("\n\t -wl for wordless");
      printf("\n\t -a for detailed display");
      printf("\n");
      return 1;
    }
  double mink=2.0;
  bool mixed=false;
  int argstrpos=0;
  bool wordless=false;
  bool norm=true;
  bool flagDisplay=false;
  double * numc=NULL;
  int nbc=0,nbcol=0,nbex=0;
  char blank=' ';
  double **diss=NULL;

  if(argc > 3) 
    {
      for(int i = 3; i < argc; i++)
        {
	  if(! strcmp("-c", argv[i])) 
	    {
	      mixed=true;
	      //next argument will hold string of combinations Euclidean and FP-based
	      argstrpos=i+1;
	    }
	  if(! strncmp("-un", argv[i],2)) 
	      norm=false;

	  if(! strncmp("-m", argv[i], 2)) 
	    {
	      mink=strtod(argv[i] + 2, NULL);
	      if (mink<=0)
		{
		  printf("Minkowski power must be >0\n");
		  return(1);
		}
	    }
	  if(! strcmp("-a", argv[i])) 
	      flagDisplay=true;


	  if(! strcmp("-wl", argv[i]))
	    wordless=true;
	}
    }
 
  // main loop for catching exceptions in called functions
  try{
    // Construction of FIS S
    FIS * S=new FIS(argv[1]);
   
    if (flagDisplay)
      printf("Minkowski power=%f\n",mink);

    // handle mixed (-c) argument
    // false = Euclidean distance, true = fuzzy partition-based distance
    
    // argstrpos=position of string
    //numc: integer array (length=number of inputs), 0 for Euclidean, 1 for FP
    nbcol=S->GetNbIn();
    numc=new double[nbcol];
    // default=fp-based fuzzy partitions in all dimensions
    for (int kk=0;kk<nbcol;kk++)
      numc[kk]=1;

    if (mixed)
    {
      int lenc=strlen(argv[argstrpos]);
      // one element: shortcut for identical choices in all dimensions
      if (lenc==1)
	{
	  numc[0]=atol(argv[argstrpos]);
	  if ((numc[0]!=0)&&(numc[0]!=1))
	    {
	      printf("\nerror in c string\n");
	      printf("only 0 or 1 are allowed");
	      return 1;
	    }
	  if (!wordless)
	    {
	      if (numc[0]==0)
		printf("Euclidean distance in  all dimensions\n");
	      else
		printf("FP-based distance in  all dimensions\n");
	    }
	  for (int pp=1;pp<nbcol;pp++)
	    numc[pp]=numc[0];
	}
      else
	{
	  nbc=SearchNb(argv[argstrpos],numc,nbcol,blank);
	
	  if (flagDisplay)
	    printf("%d choices found (0 for Euclidean, 1 for FP-based): %s\n",nbc,argv[argstrpos]);
	  
	  if ((nbc>0)&&(nbc<nbcol))
	    {
	      printf("\nerror: not enough elements in c string\n");
	    }
	  for (int kk=0;kk<nbc;kk++)
	    {
	      if ((numc[kk]!=0)&&(numc[kk]!=1))
		{
		  printf("\nerror in c string\n");
		  printf("only 0 or 1 are allowed\n");
		  return 1;
		}
	    }
	}
    }
  if(!wordless)
    {
      printf("Distance combination with Minkowski power=%g\n",mink);
      if(norm) 
	printf("Normalized distance\n");
      else
	printf("Unnormalized distance\n");
	  
      for (int p=0;p<nbcol;p++)
	{
	  if (S->In[p]->IsActive())
	    {
	      if (numc[p]==0)
		printf("column %d, Euclidean\n",p+1);
	      else
		printf("column %d, FP-based\n",p+1);
	    }
	}
    }
  // call dist on current FIS and current data file
    // 
    diss=S->dist(argv[2],nbex,(char*)"diss.out",numc,norm,mink,flagDisplay,wordless);

    delete [] numc;

    // delete diss array
   // delete diss array
  if (diss!=NULL)
    {
      for( int i=0 ; i<nbex; i++ )
	delete [] diss[i];
      delete [] diss;
      diss=NULL;
    }
  
    // delete global variable
    if(VarNameG != NULL) 
      {
	for(int j = 0; j < NbVarG; j++)
	  if(VarNameG[j] != NULL)  delete [] VarNameG[j];
	delete [] VarNameG;
	VarNameG = NULL;
      }
    if (S !=NULL)
      delete S;
    return (0);
  }
  //action in case of exception
  catch(std::exception &e){
    printf("error in dist\n");
    printf("catch %s\n",e.what());
    
    return(-1);
  }
}


//**************************   End of PERF.CPP   *************************

