//*********************************************************************
//
//
//                              CALCTREE.CPP
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISTREE classes for fuzzy decision tree
//
//*******************************************************************
//
//
//                FISTREE functions to perform node calculations
//                FISTREE is a derived class of FIS
//
//*******************************************************************

#include "fistree.h"

//-----------------------------------------------------------------------------------------------
// NODE class functions
//-----------------------------------------------------------------------------------------------

double NODE::CalcMuSigma2(int dimNode,int mfNode, double ** x, int nbex, int dataOutputCol, double mumin, 
			 FISIN ** in, char * conj, double &sigma2, double&mucum, int &count)
  // calculates mean and variance (of chosen output) for items such as mu>=mumin
  // calls CalcMuN for previous node item memberships
  // combines dimNode,mfNode membership for current node
{
  int numex;
  double average=0.0,munPrev=0.0,munMF,munNode;
  
  sigma2=0.;
  mucum=0.0;
  count=0;
  for (numex=0;numex<nbex;numex++) //item loop
    {
      munPrev=CalcMuN(x,numex,in,conj); // node item membership
      // munNode: recursive calculation
      // from current node up to root
      if (dimNode == -1)
	munMF=1.;//neutral element for root
      else
	munMF=in[dimNode]->GetADeg(mfNode,x[numex][dimNode]);
      munNode=AND(munMF,munPrev,conj);


      if(munNode > mumin) //membership is significant
	{
	  count++; // attracted examples
	  mucum=mucum+munNode;
	  average+=munNode*x[numex][dataOutputCol]; //output value
	}
    }
  // there are some examples attracted by node
  if(mucum > EPSILON)
    {
      average=average/mucum;
  
      for (numex=0;numex<nbex;numex++)
	{
	  munPrev=CalcMuN(x,numex,in,conj); // item node membership
	  if (dimNode == -1)
	    munMF=1.;//neutral element for root
	  else
	    munMF=in[dimNode]->GetADeg(mfNode,x[numex][dimNode]);
	  munNode=AND(munMF,munPrev,conj);

	  if(munNode > mumin) //membership is significant
	    {
	      sigma2+=munNode*(x[numex][dataOutputCol]-average)*(x[numex][dataOutputCol]-average);
	    }
	}
      
      sigma2=sigma2/mucum;
  
    }
  // otherwise all returned values are zero
    
  return(average);
	      
}

double NODE::CalcMuN(double ** x, int numx, FISIN ** in, char * conj)
{
  
  double muprec,mu=1.;
  NODE *wnode;
  int numvari, numMFi; // current var and MF for going up the tree
  
  //computes mu at node, depending on previous nodes in tree branch
  wnode=this;
  try
    {
      //go upward till root
      while (wnode !=NULL)
	{
	  // returns node's numvar and numMF
	  numvari=(*wnode).GetVar();numMFi=(*wnode).GetMF();
	  if (numvari >=0 && numMFi >=0)
	    {
	      //computes membership to current fuzzy set 
	      muprec=in[numvari]->GetADeg(numMFi,x[numx][numvari]);
	      //performs AND operation of current mumbership and previous one
	      mu=AND(mu,muprec,conj);
	    }
	  //updates current node with father node
	  wnode=(*wnode).GetFather();
	}
      return(mu);
    }
  catch(...)
    {
      throw;
    }
}


int NODE::PerfClassif(double ** x, int nbex, int &counter, 
		      FISTREE * fistree, double mumin, double muthresh,
		      int flagDisplay)
{
  // classif or not classif :
  //modifies counter= number of examples attracted by the node
  //if classif computes number of misclassified
  //only considers examples with mu >mumin
  //misclassified when mumax for class different from node majority class (MajClass)
  //in case of ambiguity threshold is used...
  //
  int mis=0,nc0,numex,numclas,classif=0;
  int outputNumber,dataOutputCol;
  double ex,mu,muex,muex0,muex1;
  FISIN ** in;
  FISOUT ** out;
  int nbclas;

  // get info from tree
  in=(*fistree).In;
  out=(*fistree).Out;
  outputNumber=(*fistree).GetOutputNumber();
  dataOutputCol=outputNumber+ ((*fistree).GetNbIn());
  nbclas=(*out[outputNumber]).GetNbMf();
  classif =(*fistree).GetClassif();
  counter=0;
  //
  
  
  for (numex=0;numex<nbex;numex++) //item loop
    {
      mu=CalcMuN(x,numex,in,fistree ->TypeConj() ); // item node membership
      if(mu > mumin) // significant membership
	{
	  counter++; // regression or classification - # attracted examples
	  if (classif)
	    {
	      // examining classes = finding majority class# for each item
	      //muex1=mu(item to node maj.class)
	      //muex0=max(mu) item
	      muex0=0.;muex1=0.;nc0=0;
	      
	      ex=x[numex][dataOutputCol]; //value de la sortie
	      for (numclas=0;numclas<nbclas;numclas++)
		{
		  muex=out[outputNumber]->GetADeg(numclas,ex);
		  if (numclas==MajClass) muex1=muex;
		  if (muex>muex0) 
		    {
		      muex0=muex;
		      nc0=numclas; //majority class of ex
		    }
		} //end class loop 
	      if (nc0 != MajClass) 
		{
		  // test to check if node majority class item membership is very close to
		  // membership max muex0, in this case item is well classified by node
		  // else it is misclassified
		  if (fabs(muex1-muex0)>muthresh)
		    mis++;
		}
	      // printing mu(x) - classification case
	      /*if (flagDisplay)
		printf("\n Item %d \t mu(Node)=%8.3f \t  maj.class  observed output %d, memb. = %g",
		       numex+1,mu,nc0+1,muex0);
	      */
	    } //end of classif case
	} //end loop on signif mu
    }// end on item loop
  return mis;
} //end PerfClassif

void FISTREE::UpdateDim(int nbclas,int nbMF, int * clasmaj, double * entroMF,double * entroMF0,
			double ** pnDim,double * pnDimTot,double * pnDimTot0,double ** propDimTot0,
			double * muCum0,int * leaf,int * supnode)
{
  //
  // entropy is lower, this dimension is better than the previous one
  // modifies for each MF: pnDimTot0= sum of mu for maj. class, propDimTot0=prop. of mu for each class, muCum0=sum of mu
  //                       leaf and supnode
  int mf,numclas;
  double muTemp;

  for (mf=0;mf<nbMF;mf++)
    {
      entroMF0[mf]=entroMF[mf];
      muCum0[mf]=pnDimTot[mf]; // pnDimTot will be reused for maj.class.prop
		      
		      
      // determination of majority class and of fuzzy prop.
      // for corresponding items
      // majority class fuzzy prop. is stored
      muTemp=0.;
      for (numclas=0;numclas<nbclas;numclas++)
	{
	  //
	  if (pnDim[numclas][mf]>muTemp) 
	    {
	      muTemp=pnDim[numclas][mf];
	      clasmaj[mf]=numclas;
	      // to store for each MF
	    }
	}
      pnDimTot0[mf]=muTemp; // prop. class maj. in wnode

      for (numclas=0;numclas<nbclas;numclas++)
	if (fabs(pnDimTot[mf]) >EPSILON) 
	  propDimTot0[mf][numclas]=pnDim[numclas][mf]/pnDimTot[mf];

		      
    }
		  	      
  // test for each child MF if it is a leaf
  // i.e. if all items have the same class
  // case if PnDim[numclas][MF]<= PnDimTot[MF] for one class
  for (mf=0;mf<nbMF;mf++)
    {
      leaf[mf]=false;
      for (numclas=0;numclas<nbclas;numclas++)
	{
	  if (fabs(pnDim[numclas][mf]-pnDimTot[mf]) <=EPSILON) 
	    leaf[mf]=true;
	}
    }
		
		  
  // test for each MF if it is useless 
  //(all items have a zero membership)
  for (mf=0;mf<nbMF;mf++)
    {
      if (fabs(pnDimTot[mf]) <=EPSILON) 
	supnode[mf]=true;
      else
	supnode[mf]=false;
    }
} // End UpdateDim

int FISTREE::SelectDimRelGain(double * Tgain,NODE * wnode,int * var,int dimr, 
			      double * entroMF,double ** pnDim,double * pnDimTot,
			      int flagDisplay=0)
{
  //
  int ki,nbMF=0,dim,dim0=-1,mf;
  double averGain=0.0,IV=0.0,gainRel,gainRel0,enDim,pnT; // mean of absolute gain
  
  //compute mean of Tgain
  averGain = calcAver(Tgain, dimr);
  if (flagDisplay) {
    printf("\n\n");
    for(ki=0;ki<dimr;ki++) printf(" Tgain[%d]=%f", ki, Tgain[ki]);
    printf("\t\taverGain = %f\n", averGain);
  }
		
  gainRel0 = 0.0;
  gainRel = 0.0;

  for(ki=0;ki<dimr;ki++)
    {
      if (Tgain[ki]>averGain)
	{
	  IV = 0.0;
	  dim = var[ki];
	  nbMF=In[dim]->GetNbMf();
	  enDim=Entropy(wnode,dim,nbMF,pnDim,pnDimTot,pnT,entroMF,flagDisplay);//updates pnDim pnDimTot pnT entroMF
	  if (pnT > EPSILON)
	    {
	      for(mf = 0; mf< nbMF; mf++)
		{
		if (fabs(pnDimTot[mf]) > EPSILON)
		  IV += (pnDimTot[mf]/pnT)*log(pnDimTot[mf]/pnT)/LOG2;
		}
	    }
	  IV = -IV;
	  if (IV > EPSILON) gainRel = Tgain[ki]/IV;
	  if (flagDisplay)
	    {
	      printf("\n Var=%d\t Entropy= %f\t IV = %f\t", ki,enDim,IV);
	      printf("gainRel = %f\n", gainRel);
	    }
	  if (gainRel>=gainRel0)
	    {
	      gainRel0 = gainRel;
	      dim0 = dim;
	    }
	} // only for variables above average absolute gain
    }
  if (gainRel0>EPSILON) 
    {
      // update arrays entroMF pnDim pnDimTot=ONE LAST CALL TO ENTROPY
      enDim=Entropy(wnode,dim0,nbMF,pnDim,pnDimTot,pnT,entroMF,flagDisplay);//updates pnDim pnDimTot pnT entroMF
      // update arrays: clasmaj leaf supnode entroMF0 pnDimTot0 propDimTot0 will be done in calling function
    }
  return(dim0);; // if no added info. value - keep variable found by absolute gain dim0=-1
}
int FISTREE::ChooseDim(NODE * wnode,
		       double &enDim0, int * clasmaj,  
		       double * pnDimTot0, double ** propDimTot0,double * entroMF0, 
		       double * muCum0,
		       int * leaf, int * supnode, 
		       double ** pnDim, double * pnDimTot, double * entroMF, 
		       int grel,
		       int flagDisplay
		       )
{
  //------------------------------------------------------------------------------------------------------------------------
  // examples = ptr on data array, nbl =number of rows in data file
  // in = ptr on FIS input, out = ptr on FIS output
  // nbclas = output class#
  // mumin = minimum membership for an itme to be counted in the node
  // muthresh = membership for classes

  // calculates the split gain for each remaining variable
  // if mu_nDim=0 for all items in a node, the node must be removed
  // if EnDim=0 the node is a leaf (all items have the same class)
  
  // return arguments are in
  // EnDim0,clasmaj,PnDimTot0,PropDimTot0,EntroMF0,leaf,supnode,
  // other arguments PnDim,PnDimTot,PnDimTot are working arrays allocated
  // once in Tree and used at each call of ChooseDim
  // classification tree: FISTREE::Classif=1, entropy criterion
  
  //------------------------------------------------------------------------------------------------------------------------

  int k,mf,ki,dim,dim0,dimr,dimgrel,numclas,nbMF,nbMFmax;
  int nbclas,supdim,nbl;
  int * var;
  double gain,gain0;;
  double prop,pnTfather,pnT;
  double mumin,muthresh,minlcard;
  double enBefore, enDim;
  double * Tgain=NULL; // array of absolute gain
  double ** examples;
  FISIN ** in;
  FISOUT ** out;

  //--------------------------------------------------------------------------------
  //-----------initialization-------------------------------------------------------
  //--------------------------------------------------------------------------------
  // get infos from tree object
  in=(*this).In;
  out=(*this).Out;
  nbclas=(*out[OutputN]).GetNbMf();
  examples=(*this).Examples;
  nbl=(*this).Snbrow;
  mumin=(*this).Smumin;
  muthresh=(*this).Smuthresh;
  minlcard=(*this).Sminleafcard;
  
  //print  infos on node to split
  if(flagDisplay) printf("\nInput in ChooseDim, node to split:\n");
  if(flagDisplay) (*wnode).Print(examples,nbl,this,mumin,muthresh,0,0);

  
  // select maximum entropy variable
  // test all remaining variables in node 
  // get node info
  dimr=(*wnode).GetTLDim();
  var = (*wnode).GetListDim();
  enBefore=(*wnode).GetEn(); //node field
  // variable init
  dim0=var[0];
  gain0=0.0;
  supdim=false;
      
  //determination of the maximum MF# for examined variables
  nbMFmax=0;
  for (k=0;k<dimr;k++)
    {
      dim=var[k];
      nbMF=in[dim]->GetNbMf();
      if(nbMFmax>nbMF)nbMFmax=nbMF;
    }
      

  // init arrays with maximum MF
  for (mf=0;mf<nbMFmax;mf++)
    {
      pnDimTot0[mf]=0.0;
      muCum0[mf]=0.0;
      for (k=0;k<nbclas;k++)
	propDimTot0[mf][k]=0.0;
      entroMF0[mf]=0.0;
      clasmaj[mf]=-1;
    }

  //alloc and nitialization of TgainRel to 0.0
  if (grel == 1) Tgain = Alloc1DDoubleWorkingArray(dimr);

  //--------------------------------------------------------------------------------
  //-----------initialization is finished-------------------------------------------------------
  //-------------------------------------------------------------------------------- 
  
  //-------------------examine each variable in turn
  //------------------------------------------------
  //
  //error handling try and catch
  try
    {
      if (Classif)
	{
	  // begin loop on variables
	  for (ki=0;ki<dimr;ki++)
	    {
	      dim=var[ki]; //assigning real input number
	  
	      //get MF# for studied variable
	      nbMF=in[dim]->GetNbMf();
	      // array alloc done in calling function
	      // array init for each variable
	      if (Classif)
		{
		  enDim=0.0;
		  prop=0.0;
		  pnT=0.0;
		  pnTfather=0.0;
	  

		  if (flagDisplay) printf("\n\tExamining dim=%d\n",dim);

		  // loop on all MF's for the current variable-compute entropy for each MF
		  // partial stand. and sum for computing new entropy
		  // call Entropy function or Deviance function
		  //----------------------------------------------------------
	  
		  enDim=Entropy(wnode,dim,nbMF,
				pnDim,pnDimTot,pnT,entroMF,flagDisplay);
	  
		  // goal : Entropy should go down and be as small as possible
		  //---------------------------------------------------------

		  // if PnT=0 no  active items for this dimension
		  // pass to next dim
		  // be careful if true for all dims
		  // the node should not be built(flag supdim)
		  if (pnT <=EPSILON)
		    {
		      if (flagDisplay) printf("\nWARNING: in choosedim if dim=%d\tEnBefore=%g\tPnT=%g",dim,enBefore,pnT);
		      supdim=true;
		    }
		  else
		    {
		      supdim=false;
		      //standardize with MF memb.sum
		      enDim/=pnT;
		      // compute absolute entropy gain
		      gain=enBefore-enDim;
		      // if relative gain option store gain
		      if (grel==1) Tgain[ki] = gain;
		      //
		      if (flagDisplay) //detailed printing
			{
			  printf("\nin choosedim if dim=%d\tEnBefore=%g\tEnDim=%g\tGain=%g\tPnTfather=%g\tPnT=%g",
				 dim,enBefore,enDim,gain,pnTfather,pnT);
			  for (k=0;k<nbMF;k++)
			    {
			      printf("\nk=%d\tEntroMF[%d]=%g\tPnDimTot[%d]=%g\n",k,k,entroMF[k],k,pnDimTot[k] );
			      for (numclas=0;numclas<nbclas;numclas++)
				printf("\tPnDim[%d][%d]=%g",numclas,k,pnDim[numclas][k]);
			    }
			  printf("\n");
			}
	      
		      if (ki==0) enDim0=enDim;
	      
		      if(enDim<=enDim0)
			{
			  dim0=dim;
			  gain0=gain;
			  enDim0=enDim;
			  // update arrays: clasmaj leaf supnode entroMF0 pnDimTot0 propDimTot0
			  UpdateDim(nbclas,nbMF,clasmaj,entroMF,entroMF0,pnDim,pnDimTot,pnDimTot0,propDimTot0,muCum0,leaf,supnode);
			}
		    } // end of else on PnT
		} // end classification case
	    } // end of loop on variables
	
      
	  // final treatment: 2 cases - 0 absolute gain, 1 relative gain - one more loop
	  // if entropy went up,or if gain is too small, node should not be split
	  // case grel =0: absolute entropy gain
     
	  // case grel=1: relative entropy gain- compute IV=information value of each variable
	  if (grel == 1)
	    {
	      dimgrel=SelectDimRelGain(Tgain,wnode,var,dimr,entroMF,pnDim,pnDimTot,flagDisplay); 
	      // returns -1 if no better dimension found (keep abs. gain dimension), dim0=dimension of max.rel.ent.gain  otherwise
	      if (dimgrel >=0)
		{
		  // re-update arrays: clasmaj leaf supnode entroMF0 pnDimTot0 propDimTot0
		  nbMF=in[dimgrel]->GetNbMf();
		  UpdateDim(nbclas,nbMF,clasmaj,entroMF,entroMF0,pnDim,pnDimTot,pnDimTot0,propDimTot0,muCum0,leaf,supnode);
		  dim0=dimgrel;
		}
	    }
	  if ( (enBefore-enDim0)<EPSILON )
	    {
	      supdim=1;
	      enDim0=enBefore; //no entropy change
	      dim0=-1;
	    }
	} // end classification case final treatment	
      return(dim0);
    }//end of try

  catch(...)
    {
      throw;
    } 
} // end of choosedim function

int FISTREE::ChooseDimReg(NODE * wnode, double & sigma2Zero,
			  double * ave0, double *sig2Zero, double * mucum0, 
			  int * leafk0, int * supk0, 
			  double * ave, double *sig2, double * mucum,
			  int flagDisplay)
{
  // chooses variable for split of regression tree
  // regression tree:     FISTREE::Classif=0, deviance criterion
  // returns chosen dimension for split
  // modifies its arguments: ave0-sig20-mucum0-supk0
  // uses working arrays ave-sig-mucum
  int ki,mf,dim,dim0,nbMF,dimr,count=0;
  int * var;
  double sigmat2=0.,wsigmat2=0,sigma2Before=0.,smucum=0.;
  flagDisplay=0;
  //init
  dimr=(*wnode).GetTLDim();
  var = (*wnode).GetListDim();
  sigma2Zero=0.;
  sigma2Before=(*wnode).GetEn();
  sigma2Before=sigma2Before*sigma2Before;
  dim0=-1;

  //print  infos on node to split
  if(flagDisplay) 
    {
      printf("\nInput in ChooseDim, node to split:\n");
      (*wnode).Print(Examples,Snbrow,this,Smumin,Smuthresh,0,0);
      printf("\tsigma2=%g\n",sigma2Before);
    }

  // begin loop on variables
  for (ki=0;ki<dimr;ki++)
    {
      dim=var[ki]; //assign real input number
      if (flagDisplay) printf("\n\tIn choosedimReg examining dim=%d\n",dim);      //get MF# for studied variable
      nbMF=In[dim]->GetNbMf();
      sigmat2=0.;
      wsigmat2=0.;
      smucum=0.;
      for (mf=0;mf<nbMF;mf++)
	{
	  mucum[mf]=0.0;
	  ave[mf]=wnode->CalcMuSigma2(dim,mf,Examples,Snbrow,OutputN+NbIn,Smumin,In,TypeConj(),sig2[mf],mucum[mf],count);//sig2[mf]=MF subnode variance
	  if (flagDisplay) printf("\nIn choosedimReg if dim=%d and mf=%d,\tmucum[%d]=%g  #examples=%d\tsig2[%d]=%g\tave[%d]=%g",
				  dim,mf,mf,mucum[mf],count,mf,sig2[mf],mf,ave[mf]);
	  if (ki==0)
	    {
	      ave0[mf]=ave[mf];
	      sig2Zero[mf]=sig2[mf];
	      mucum0[mf]=mucum[mf];
	    }
	  sigmat2+=sig2[mf];
	  wsigmat2+=sig2[mf]*mucum[mf];
	  smucum+=mucum[mf];
	  
	}
      if(smucum>EPSILON)
	wsigmat2=wsigmat2/smucum;

      if (flagDisplay) printf("\nsigma2 total=%g\ttotal cumulated mus=%g\tweighted sigma2 tot.=%g\n",sigmat2,smucum,wsigmat2);
      if (ki==0)
	{
	  dim0=dim;
	  sigma2Zero=wsigmat2;
	  
	}
      if (wsigmat2 < sigma2Zero)
	{
	  dim0=dim;
	  //sigma2Zero=sigmat2; //modif ponderation
	  sigma2Zero=wsigmat2;
	  for (mf=0;mf<nbMF;mf++)
	    {
	      ave0[mf]=ave[mf];
	      sig2Zero[mf]=sig2[mf];
	      mucum0[mf]=mucum[mf];
	    }
	}
		
    }
  if (sigma2Zero > sigma2Before) // ponderation
    dim0=-1;
  else
    {
      //update
      nbMF=In[dim0]->GetNbMf();
      for (mf=0;mf<nbMF;mf++)
	    {
	      supk0[mf]=0;
	      //not enough examples - subnode will not be created
	      if (mucum0[mf] < EPSILON)
		supk0[mf]=1;
	      // subnode is pure (std=0)
	      leafk0[mf]=0;
	      if (sig2Zero[mf] < EPSILON)
		leafk0[mf]=1;
	      
	    }
    }

  return(dim0);

}// end of choosedimreg function
