//*********************************************************************
//
//
//                              FISTREE.H
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISTREE classes for fuzzy decision tree
//*******************************************************

//******************************************************************
//
//                  DEFINITION OF NECESSARY CLASSES 
//                  FOR THE FUZZY DECISION TREE
//
//
//******************************************************************
//******************************************************************
//
//                        NODE CLASS
// 
//
//******************************************************************
#include "fis.h"
#include <math.h>
#define LOG2 log(2.)
#define MAXCLASSES 99
//************************************************************************************************************************

class FISTREE;
//! a node is a fuzzy tree element
class NODE
//**********
{  
  NODE& operator=(const NODE&);
  NODE(const NODE&);
  
  protected :
  //! node number starting at zero for root
  int OrderNum;
  //!node characteristic input variable number
  int Numvar;
  //!node characteristic MF number
  int NumMF;
  //! Leaf=1:  The node is a leaf node
  int Leaf;
  ///! LeafPrev=1:  The node was a leaf node before pruning
  int LeafPrev;
  //!majority fuzzy proportion (classif.) or average value (reg.)
  double Propn;
  //! fuzzy proportion for each class
  double * PnClas; 
  //! node majority class
  int MajClass; 
  //! node entropy after split (classif.) or std (reg.)
  double En; 
  //! cumulated membership for all examples attracted by node
  double MuCum;
  //! number of children of this node
  int NChildren; 
  //! number of children of this node before pruning
  int NChildrenPrev; 
  //! current child
  int NumChildC; 
  //! ptr on child node array
  NODE ** Child; 
  //! ptr on childprev node array
  NODE ** ChildPrev;
  //! ptr on father node (only one)
  NODE * Father; 
  //! integer array= remaining variables indices
  int *ListDim; 
  //! its size
  int ListDimSize; 
  //! flag for bottom up algorithm
  int PrExam;

 
  
  //! removes the variable VarRem from ListDim array
  void UpdateListDim(int * listeVarAv,int tailleAv, int varRem,int * listeVarAp ); 

  //! calculates the N node item membership
  double * MuN(double ** ex, MF * pMF); 

  
 public:

  //! static field: holds the number of generated nodes
  static int StatnbNode;

  //! Unused constructor.
  NODE(){};

  //! NODE constructor
  //! allocates PnClas and ListDim arrays
   
  NODE(int classif,int dim, int nMF, int leaf,int clmaj,
       double pnode, double * propnode, double enode, double mucum,
       NODE * upnode,  int nbcl, int nbChildnode=0,
       int numVarToRemove=-1)
    {
      int i;
      //
      // node info
      StatnbNode++;
      OrderNum=StatnbNode;
      Numvar=dim;
      NumMF=nMF;
      PrExam=0;
      
      Leaf=leaf;
      LeafPrev=Leaf;
      MuCum=mucum; // cumulated membership
      Propn=pnode;  // holds majority fuzzy proportion (classif.) or average value (reg.)
      if (classif)
	En=enode; //! node entropy after split (classif.) or std (reg.)
      else
	{
	  if(enode>EPSILON)
	    En=sqrt(enode); //std is stored instead of variance
	  else
	    En=0.;
	}
      NChildren=nbChildnode;
      NChildrenPrev=NChildren;
      PnClas=NULL;

      if (classif) // begin classif case
	{
	  MajClass=clmaj;
	  Propn=0.;
	  //alloc class fuzzy prop array
	  if (nbcl>0)
	    {
	      PnClas = new double[nbcl];
	      for (i=0;i<nbcl;i++)
		{
		  PnClas[i]=propnode[i];
		  if (PnClas[i] >Propn)
		    Propn=PnClas[i];
		}
	    }
	} //end classif case
     
      // the father node already exists
      Father=upnode;
      NumChildC=0;
     
      Child=NULL;
      ChildPrev=NULL;
      ListDim=NULL;
      // alloc array of child nodes and childprev nodes
      if (NChildren > 0) 
	{
	  printf("\ncreating a size %d array for child node ptrs\n",NChildren);
	
	  Child = new NODE *[NChildren];
	  ChildPrev = new NODE *[NChildren];
	}
      // list available  variables to build the node children
      // alloc  ptr
      // array alloc is done here
      ListDimSize=0;
      if (Father != NULL)
	{
	  if (numVarToRemove == -1)
	    ListDimSize=(*Father).ListDimSize;
	  else
	    {
	      ListDimSize=(*Father).ListDimSize - 1;
	    }
	
	  if (ListDimSize >=1)
	    {
	      ListDim=new int [ListDimSize];
	      UpdateListDim((*Father).ListDim,(*Father).ListDimSize,numVarToRemove,ListDim);
	    }  
	  else
	    {// no remaining variable, node is a leaf
	      SetLeaf(1);
	    }
	}
      // else root node, holds all variables

     
    }
  //! to DELETE a node, all its children must first be deleted,  
  //! then ptr array Child can be deleted, 
  //! Be careful to call this destructor only after all child nodes have been deleted
  virtual ~NODE() 
    { 
      // destroy dynamic arrays
      //----------------------------------------------------------------------------
      // IMPORTANT:  called by UpDownTree
      //----------------------------------------------------------------------------
     
      if (this != NULL)
	{
	  //printf("deleting node %d\n",OrderNum);
	
	  if (ListDim != NULL) 
	    {
	      delete [] ListDim;
	      ListDim=NULL;
	    }

	  if (PnClas != NULL)
	    {
	      delete [] PnClas;
	      PnClas=NULL;
	    }

	  if (Child != NULL) 
	    {
	      delete [] Child; 
	      Child=NULL;
	    }
	  
	  if (ChildPrev != NULL) 
	    {
	      delete [] ChildPrev; 
	      ChildPrev=NULL;
	    }
     	}
      
    }//end destructor


  
  
 

  double AND(double mu1, double mu2, char * conj); //!< AND operator

  double IMPLI(double mu1,double mus,char * conj); //!< Implication operator

  // Access to private data
  
  //! returns the number of children
  int GetNChildren() const;
  //! returns the number of children before pruning
  
  int GetNChildrenPrev() const;

  //! sets the number of children
  void SetNChildren(int nbc){NChildren=nbc;}

  //! returns the current child node number
  int GetNumChildC() const;
 
  //! updates the current child number
  void SetNumChildC(int num); 

  //! returns the remaining variable list for this node
  int* GetListDim();

  //! returns the size of the list
  int GetTLDim() const;

  //!returns 1 if node is a leaf, 0 otherwise
  int GetLeaf() const;

  //!returns 1 if node was a leaf before pruning, 0 otherwise
  int GetLeafPrev() const;

  //! sets Leaf field to 1
  void SetLeaf(int f);

  //! sets LeafPrev field to 1
  void SetLeafPrev(int f);

  //! returns the node number
  int GetOrderNum() const;

  //! returns the node characteristic variable number
  int GetVar() const;
 
  //! returns the node characteristic MF number
  int GetMF() const; 

  //! returns the node majority class
  int GetMajClass() const;

  //! returns the node average value or Propn
  double GetAverageValue() const {return(Propn);}

  //! sets the MajClass field to cl
  void SetMajClass(int cl);

  //! returns the node entropy En
  double GetEn() const;

  //! sets the ListeDim array field to listevar
  void SetListDim(int * listevar, int nbvar);

  //! returns a node ptr on the node father
  NODE * GetFather();

  //! returns a node ptr on the child node number numf
  NODE * GetChild(int numf);

  //! returns a node ptr on the child node number numf before pruning
  NODE * GetChildPrev(int numf);

  //! removes the child node numf from the Child array, and deletes it
  void RemoveChildNode(int numf, int flagDisplay=0); //updates node child info
  
  //! removes all subnodes
  void RemoveAllChildren(int flagDisplay=0);

  //! allocates the Child ptrs with the given size
  void CreateChildren(int nch);

  //! transfers the  node ptr in node.Child[numf], 
  //! ptr must exist
  
  void SetChild(int numf, NODE * ptr); 

  void SetPrExam(int prexam){PrExam=prexam;}

  int GetPrExam() const {return(PrExam);}


  /*! calculates mean and std for items such as mu>=mumin, 
    \param argument sigma modified 
    \return std
  */
  
  
  double CalcMuSigma2(int dim, int mf, double ** x, int nbex, int numvars, double mumin, 
		     FISIN ** in, char * conj, double &sigma2, double &mucum, int & count);

  /*! returns mu(item x) for this node,  
    performs an iterative calculation of mu(x[numx][numvar] for MF NumMF),  
    for all previous nodes in tree branch
    \param conjunction = determines AND function for combining successive mu
  */ 
 
  double CalcMuN(double ** x, int numx, FISIN ** in, char * conjunction);

  //! prints node characteristics
  void Print(double ** x, int nbex, FISTREE * s,
	     double mumin, double muthresh, int display, int depth=0, FILE *f=stdout);

  //! prints tree summary in file f, 
  //! calls PerfClassif and calcMuSigma, 
  //! datasep is the data separator defined in the Interface section of FIS config file, 
  void PrintVisu(double ** x, int nbex, FISTREE * s,
		 double mumin, double muthresh, char datasep,FILE *f=stdout);

  //! classif or not classif:  
  //! in both cases modifies nbtot argument, 
  //! = number of examples attracted by the node, 
  //! only considers examples with mu >mumin, 
  //! if classif returns number of misclassified, 
  //misclassified when mumax for class different from node majority class (MajClass), 
  //!in case of ambiguity, threshold muthresh is used...
  int PerfClassif(double ** x, int nbex, int & nbtot, FISTREE * s, double mum, double mus,
		  int display); //calculates the classification performance
  
  //!updates the node En field after splitting
  void SetEn(double EnAp);

  //!updates the node Propn field after splitting
  void SetPropn(double propn);

  //! sets mucum
  void SetMuCum(double mucum){MuCum=mucum;}

  //!updates the node PnClas and Propn fields after splitting
  void SetProps(double * propnode, int nbcl);

  //! Naive criterion pruning function, 
  //! obsolete and replaced by PrTreePerf
  int NaiveCriterion(double ** examples,  int nbr, FISTREE * s, double mum, double mus, int display);
  
  //! analyze node and update fistree info
  void Analyze(double ** x, int nbex, FISTREE * fistree,
	       double mumin, double muthresh,int flagDisplay);

};// end of NODE class

//*******************************************************************
//
//
//                          FISTREE CLASS
//
//
//*******************************************************************


class FISTREE:public FIS
//*******
{
  FISTREE& operator=(const FISTREE&);
  FISTREE(const FISTREE&);
 protected:  
  NODE * Sroot; //!< ptr on the root node
  int OutputN; //!< output Number for building tree
  int Classif; //!< 1-classification case, 0-regression case
  // classif.case if fuzzy output or classif=yes
  // regression case if crisp output and classif=no
  int Snbclass; //!< number of output classes
  int Snbrow;//!< data sample number of rows
  int Snbcol;//!< data sample number of columns
  int Snbleaf; //!< tree leaf#
  double Smumin; //!< minimum significant membership
  double Smuthresh; //!< threshold membership tolerance for a class
  int Sminleafcard; //!< minimum number of items for further splitting a node
  double ** Examples;//!< ptr on data
  int * NodeRule;//!< array for node-rule correspondence
  int LnodeRule;//< previous array length
  

   
  //! sets Sroot to ptr argument
  void SetRoot(NODE * ptrnode);

 public:
  // summary info
  int nbAtItems;
  int nbAtItemsMax;
  double mucumAtItems;
  double  mucumAtItemsMax;
  double stdED;
  int * depthVar;
  int * countVar;
  double CovThresh;

  FISTREE(){};

  //! constructor
  //! Construct FISTREE S from config file and data file
  //! with a ptr on root node
  //! warning!!!
  //! S will be built without rules 
  //! (even if rules are given in config file argv[1])
  //! to use it as a sif after rules will be added, initpossibles must be called.
  //! tree is based on inputs and first output
  //! output must have classes
  //! for fuzzy output, classes are MF labels
  //! for crisp output, classes are taken from data
  //! implemented with trueClasses=1 - exact class values, no tolerance
  FISTREE(char *fisConfigFile,char *dataFile, double mumin,double muthresh, 
	  int outputNumber, double *& outputClasses, FISOUT *& crispout,double covT, int minleafcard=0)
	  
    { 
      int i,crisp,trueClasses=1;
      double *dataCol;
      FISOUT *fuzzyout;

      Snbclass=0;
      Sroot=NULL; //!< init ptr on the root node that will be built later
      Smumin=mumin;
      Smuthresh=muthresh;
      Sminleafcard=minleafcard;
      NodeRule=NULL;
      LnodeRule=0;
      Examples=NULL;
      crispout=NULL;
      fuzzyout=NULL;
      dataCol=NULL;
      outputClasses=NULL;
      OutputN=outputNumber;
      // init summary info
      nbAtItems=0;
      nbAtItemsMax=0;
      mucumAtItems=0;
      mucumAtItemsMax=0;
      stdED=0.;
      CovThresh=covT;
      
     
      //! FIS exception handling
      try{
	//! remember that FIS() calls init();
	InitSystem(fisConfigFile);//InitSystem can throw an exception
	//! to avoid problem when NbRules different from zero
	//! remove rules and set NbRules to zero
	if (Rule != NULL) 
	  {
	    for( i=0 ; i<NbRules ; i++) {
	      if (Rule[i] != NULL) delete Rule[i];
	    }
	    delete [] Rule;
	  }
	Out[outputNumber]->DeletePossibles(NbRules);//delete MFConcs...
	NbRules=0;
	Rule=NULL;


	//!read sample file, set Snbcol and Snbrow fields
	Examples = ReadSampleFile(dataFile,  Snbcol, Snbrow);

	//! compute minleafcard vs number of examples
	if (minleafcard<=0)
	  Sminleafcard=(int)min(10,Snbrow/10);
	
	
	// allocate depthVar & countVar - init to 0
	depthVar=Alloc1DIntWorkingArray(NbIn,0);
	countVar=Alloc1DIntWorkingArray(NbIn,0);
	

	// check for #variables in FIS=#columns in data set
	if ((NbIn+NbOut)!=Snbcol) throw std::runtime_error("wrong~number~of~columns~in~data~file\n");
	// check for valid output number - must correspond to a column in data array

	if(((NbIn+outputNumber)>=Snbcol ) || (outputNumber <0)|| (outputNumber>(NbOut-1))) throw std::runtime_error("~error~in~outputNumber\n");
	// number of classes: number of MF's for fuzzy output
	// number of classes from data file for crisp output
	//! set Snbclass field
	crisp=strcmp(Out[outputNumber]->GetOutputType(),OUT_FUZZY::OutputType());
	// classif.case if fuzzy output or classif=yes
	// regression case if crisp output and classif=no
	Classif = (!crisp) || (Out[outputNumber] -> Classification());
	if (crisp == 0)
	  {
	    Snbclass=(*Out[outputNumber]).GetNbMf(); //!< class#
	    if (Snbclass <=0) throw std::runtime_error("~fuzzy~output~with~no~MFs\n");
	  }
	else
	  if (Classif == 1)
	  {
	    // crisp output and classif=yes flag
	    // transform crisp output into fuzzy output with discrete MF's
	    // classes are taken from data column
	    // add it as last output
	    // store the original output for perf and pruning calculations
	    dataCol= new double [Snbrow];
	    for(i = 0; i < Snbrow; i ++)  dataCol[i] = Examples[i][NbIn + outputNumber];
   
	    // fuzzyout is allocated in function Crisp2DiscreteFuzClassesFromData
	    fuzzyout=Crisp2DiscreteFuzClassesFromData(dataCol,outputClasses,trueClasses,outputNumber);
	    if (trueClasses == 0) // for the future...
	      {
		// transfer datacol into Examples array
		for(i = 0; i < Snbrow; i ++)   Examples[i][NbIn + outputNumber]=dataCol[i];
	      }
	    // adds to FIS a new FISOUT by cloning fuzzyout
	    // increments NbOut by one, sets field Out[NbOut].Num=NbOut
	    AddOutput(fuzzyout); 
	    // exchange fuzzy and crisp outputs- fuzzy output comes first 
	    Out[NbOut-1]=Out[outputNumber];
	    // keep track of ptr on original crisp output
	    crispout=Out[NbOut-1];
	    Out[outputNumber]=fuzzyout;
	    
	    // delete dataCol
	    if (dataCol != NULL) delete [] dataCol;
	    //decreases NbOut by 1 as only one output is used at a given time
	    NbOut--;
	  }

      }

      catch(std::runtime_error){
	throw;
      }
    
    }



  virtual ~FISTREE() 
    {
      // FIS destructor is (implictely) called first
      int action=4,res,nbact,ii,display=0;
      //follow tree to destroy all nodes including root
      // compute number of active inputs
      nbact=0;
      for (ii=0;ii<GetNbIn();ii++)
	nbact+=In[ii] -> IsActive();
      if (Sroot != NULL) res=UpDownTree(Sroot,action,display,nbact);
      //destroy data array
      if( Examples != NULL )
	{
	  for (int i=0;i<Snbrow;i++) 
	    if (Examples[i] != NULL) delete [] Examples[i];
	  delete [] Examples;
	}
      // destroy rule node correspondence array
      if (NodeRule != NULL && LnodeRule >0) delete [] NodeRule;
      // destroy summary info arrays
      if (depthVar !=NULL) delete [] depthVar ;
      if (countVar !=NULL) delete [] countVar ;
      //implicit call to  FIS destructor
    
    }
  
  //! builds a new fuzzy output with discrete MF's
  //! MeanMax defuziffication type
  //! arguments : data column ptr, and output number
  //! if outputNumber>=0, all parameters are taken from existing corresponding output
  //! if outputNumber=-1, the output is totally new
  FISOUT * Crisp2DiscreteFuzClassesFromData(double * dataCol, double *& outputClasses, int trueClasses,
					    int outputNumber=-1);

  //! key function-constructs the fuzzy decision tree, 
  //! default:  full tree is built with max depth and absolute entropy gain 
  //! algorithm
  //! starts from root node - built there
  //! makes first split (call to ChooseDim function) and creates child nodes accordongly
  //! then sets first child node as candidate node for split and splits it (call to ChooseDim function)
  //! once leftmost subtree is finished, go up one level and split immediate neighbour node to the right
  //! tree is built until maxdep is reached (default=all variables)
  //! no further split in one branch if entropy gain < gainEntromin
  NODE * Tree(int display, int maxdep, double gainEntromin,int grel); 
  
  //! follows decision tree from given node down to given depth
  //! returns the node# of the followed tree
  /*! 
    \param action =0 : count leaf nodes and sets Snbleaf field
    \param action =1 : print nodes
    \param action =2 : generate rules corresponding to leaves
    \param action =3 : generate FIS summary
    \param action =4 : destroy all nodes (called by ~FISTREE
    used variables from FIS class In, Out, NbIn, NbOut, Rule, NbRules
  */
  int UpDownTree(NODE * node, int fonc, int display, int maxdepth, FILE * f=stdout); 
 
  //! computes tab array average
  double calcAver(double *tab, int n);

  //! key function, 
  //! selects the node split dimension, 
  //! all modified arrays are on the second and third lines, 
  /*! 
    \param EnDimNew = node entropy after split on dim
    \param clmaj,PDimTot0,EntroMF0,leaf,supnode are multidim arrays:  
    they give info on the child nodes  
    to be built when splitting the node;  
    the fourth line holds work arrays allocated (once) and initialized in the calling function (tree)
    \param grel=0: absolute entropy criterion for splitting
    \param grel=1: relative entropy criterion for splitting
  
    calls Entropy function
  */
  int ChooseDim(NODE * node, double &EnDimNew, int * clmaj, 
		double * PnDimTot0, double ** PropDimTot0, double * EntroMF0, 
		double * muCum0,
		int * leaf,int * supnode,
		double ** PnDim, double * PnDimTot, double * EntroMF,
		int grel,
		int Display=0);

  //! chooses split in regression tree
  int ChooseDimReg(NODE * wnode, double & sigmaZero,
		   double * ave0, double *sig0, double * mucum0, 
		   int * leafk0, int * supk0, 
		   double * ave, double *sig, double * mucum,
		   int flagDisplay=0);

  //! computes and returns entropy at node wnode for variable dim
  /*! modifies its pnDim, pnDimTot, pnT and entroMF arguments
    \param pnDim : size ncl*nbMF sum of mu for each class in each subnode
    \param pnDimTot, entroMF: size nbMF - sum of mu for each subnode, entropy of each subnode
    \param pnT: sum of mu for dimension
    \param particular case: dim=-1 : root node, do not compute previous mu (muprec)

    called by ChooseDim function
  */
  double Entropy(NODE * wnode,
		 int dim,int nbMF,
		 double ** pnDim, double * pnDimTot, 
		 double &pnT, double * entroMF, 
		 int flagDisplay);
  
  //! allocates 1d double array and initialize all elements to 0.
  double ** Alloc2DDoubleWorkingArray(int nrow, int ncol);

  //! allocates 2d double array and initialize all elements to 0.
  double * Alloc1DDoubleWorkingArray(int n);

  //! allocates 1d double array and initialize all elements to 0.
  int * Alloc1DIntWorkingArray(int n, int ival);
  
  //! only prints total node # and displays rules
  //! does not call FIS print function
  virtual void Print(FILE *f = stdout) const;

  //! adds FISTREE rules
  void AddRule(){}; 
  
  //! returns outputNumber
  int GetOutputNumber() const {return(OutputN);}
  //! returns Sroot node ptr
  NODE * GetRoot();
  
  //! returns Snbrow (number of examples in data file)
  int GetNbex() const ;

  //! returns Classif flag
  int GetClassif() const {return(Classif);}

  //! returns ptr on examples read from data file
  double ** GetExamples();

  //! sets ptr on examples read from validation file
  void ResetExamples(char * valFile=NULL);  

  //! returns number of output classes
  int GetNbClass() const;

  //! sets NbOut
  void SetNbOut(int n) {NbOut=n;}

  //! sets NbRules
  void SetNbRule(int nr) {NbRules=nr;}

  /*! 
    \brief Rebuilds the ruleStr string using the integer array describing the R rule
    \param ruleStr can be passed as an argument of rule constructor
    \param The array char ruleStr must be preallocated
    \param the rule length is given in ruleLen
  */
  void RuleString(int * ruleArray,  char * ruleStr, int ruleLen);

  //! constructs root node and then 
  //! constructs full tree (with given depth), and entropy/deviance gain > minEDgain
  //! according to grel (absolute or relative entropy gain);  
  //! then prints tree, counts leaves and generates fuzzy rules;  
  //! writes FIS config file cfgOut and summary file viewTree;  
  //! calls Tree and UpDownTree functions
  int BuildFullTree( char* cfgOut, char* viewTree, int MaxDepth,  double minEDgain, int grel, 
		     FISOUT * oldout, double * outputClasses, int FlagDisplay=0);
  
  /*! prunes tree according to naive criterion - pruned tree replaces full tree
    \param writes pruned FIS config file cfgOut and summary file viewTree

    //! calls PrtTree and NaiveCriterion functions
    */
  int PruneTree( char* cfgOut, char* viewTree, int FlagDisplay=0);
  
  //! follows and prunes fuzzy decision tree from starting node start.  
  //! called by PruneTree.  
  //! obsolete
  int PrTree(NODE * start, int action, int display, int maxdepth);
  
  /*! prunes tree according to performance criterion - pruned tree replaces full tree
    \param writes pruned FIS config file cfgOut and summary file viewTree

    //! calls PrTreePerf and UpDownTree functions
    \param arguments misClass and Lab must be preallocated

    see maintree.cpp for an example

    \param argument nbrem is modified and corresponds to the number of removed nodes
  */
  int PruneTreePerf(char* cfgOut, char* viewTree, char* validFile, int*& misClass, double*& lab, 
		    double muperf, double relPerfLoss,
		    int & nbrem, FISOUT * oldcrispout, int split,
		    int flagDisplay=0);
  
  /*! follows and prunes fuzzy decision tree from starting node start 
    \param maxdepth = tree depth
    \param relPerfLoss: uses perf criterion base on relative tolerated loss argument
  
    \return node count on followed tree
    //! used variables from FIS class In, Out, NbIn, NbOut, Rule, NbRules
    \param arguments misClass and Lab must be preallocated
    //! see maintree.cpp for an example
    \param action argument is reserved for further developments
    //! called by PruneTreePerf, 
    //! recursively calls TryPruning
    */
  int PrTreePerf(NODE * start, int action, int * misClass, double * lab, 
		 double muperf, double relPerfLoss, 
		 int maxdepth,int split,int flagDisplay);

  
  /*! computes new perf,  
    calls UpdateRule which updates or deactivates rule, 
    if no pruning calls BackRule which restores previous state
    \return integer : -1 if problem, 0 if no pruning was done, 1 if pruning was done
    \param modifies its last three arguments if pruning was done
    \param numc = current child number 
    \param nbch = number of children
    \param nbRemovedNodes = number of removed nodes
  */
  int TryPruning(NODE * wnode,NODE * upnode,double perfInit, 
		 double muperf, double relPerfLoss, 
		 int * misClass,double * lab,int Flagdisplay, 
		 int split, int & numc, int & nbch, int & nbRemovedNodes, int depth);

  /*! searches for rule number corresponding to node number,
    in NodeRule array (which was built by BuildFullTree function
    \return -1 if rule was not found
  */
  int FindRule(NODE * wnode);

  /*! updates rule indR corresponding to wnode argument
    \param upnode argument = wnode father
    behaviour depends on upnode nature:  
    upnode has remaining children: rule is deactivated, deac=1, 
    upnode becomes a leaf: rule is transformed, deac=0, 
    \return deac
  */
  int UpdateRule(int indR, NODE * wnode, NODE * upnode, int split, int *& subR, double * lab, int flagDisplay);

  void PruneRule(NODE * wnode,NODE * upnode, int indR, double * lab,int flagDisplay);

  void UnPruneRule(NODE * wnode,int indR, double * lab, int flagDisplay);

  //! undoes all the work done by UpdateRule
  int BackRule(int indR, NODE * wnode, NODE * upnode,int split, int * subR, double * lab,int flagDisplay );

  //! max number of MFs in FIS input variables
  int CalcMaxMf();

  //! init root node parameters
  void InitRoot(double ** w_PnDim,double * w_PnDimTot,double ** propDim,double * w_EnDim,int * clasmaj,int flagDisplay);

  //! set variable list for root node
  void  SetRootVar(int flagDisplay);

  //! create subnodes of NODE wnode for split mindim
  //! mindim=-1 no split -> no subnodes created
  //! supnode[k]=1 -> subnode k is not created, other subnodes are created
  //! leaf[k]=1 -> subnode k is pure, not to be split in the following steps
  int CreateSubNodes(int mindim,NODE * wnode,int * leaf,int * supnode, int * clasmaj,
		     double * pnDim,double ** propDim,double * enDim,double *muCum,
		     int depth,int flagDisplay);
  
  //! in tree building 
  //! searches for the next node to be split (up and right)
  //! returns it, and modifies current tree depth
  void NextNode(NODE *& wnode, int & depth,  NODE * start);

  // next 3 functions:  init tUpDownTree algorithm (2=generate rules, 3=write summary file)
  void InitUpDownTree2(int & lRule, int *& tabRule, double & concRule,char*& cRule, int flagDisplay);

  void InitUpDownTree3(FILE * fres, NODE * wnode, int flagDisplay);

  void InitUpDownTree014(int action, NODE * wnode, int depth, int flagDisplay);

  //! called by tree function - action on all nodes including leaves
  void InternalNodeAction(int action, NODE * wnode, int & numvari, int * tabRule, FILE * fres,int depth, int flagDisplay);

  //! called by tree function - action on leaves only -returns 1 if wnode=tarting point
  int LeafNodeAction(int action, NODE *& wnode, NODE * start, int & counter, int & nbleaf, int numvari, 
		      int lRule, int * tabRule, double & concRule, char * cRule, int flagDisplay);

  //! called by UpDownTree function - loop to search for leaf and action on leaf
  int NextNodePlusAction(NODE *& wnode, NODE * upnode, int & depth, NODE * start, int action,
			 int & ruleCounter, int & nbleaf, int & numvari, 
			 int lRule, int * tabRule, double & concRule, char * cRule, 
			 FILE * fres, int flagDisplay);

   //! only leaves can be deleted - calls node destructor - sets curnode to NULL after destruction
  int DeleteLeaf(NODE *& curnode, NODE * start,int flagDisplay);

  //! called by NextNodePlusAction to prepare array for rule generation
  void UpdateTabRule(NODE * wnode, int * tabRule);

  //! called by PrTreePerf function - loop to search for leaf and action on leaf
  int NextNodePlusPruneLeaf(NODE *& wnode, int & depth, NODE * start, int action,
			    double perfInit,double muperf,double relPerfLoss,
			    int *misClass,double * lab, int & nbrem,
			    int split,int flagDisplay);
  
  //! updates arrays after choosing splitting dimension (classif. case)
  //! pnDimTot0(size #mf):  maj. class for each subnode
  //! propDimTot0(size #mf x "class): fuzzy prop. of each class for each subnode
  //! mucum0(size #mf): sum of mu for each subnode 
  //! leaf(size #mf):  1 if subnode is not to be split
  //! supnode(size #mf):  1 if subnode is not to be created (mucum0 <EPS)
  void UpdateDim(int nbclas,int nbMF,int * clasmaj, double * entroMF,double * entroMF0,
		 double ** pnDim,double * pnDimTot,double * pnDimTot0,double ** propDimTot0,
		 double * mucum0,int * leaf,int * supnode);

  //! SELECTS dimension for relative gain criterion (classif. case)
  int SelectDimRelGain(double * Tgain,NODE * wnode,int * var, int dimr, 
		       double * entroMF,double ** pnDim,double * pnDimTot,
		       int flagDisplay);

  //! for summary analysis
  void AnalyzeTree(FILE* fres,int d, int flagDisplay);
  void AnalyzeTreeWriteH(FILE* fres);
  void ResetAnalyze();
  int GetSminleafcard(){return Sminleafcard;};
};

// end of FISTREE class
