//***********************************************************************
//
//
//                              FPA.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISFPA class functions used by FISPRO, 
//        part of package fpa (Fast Prototype Algorithm)
//        requires library fispro
//
//**********************************************************************

#include "fpa.h"

double * WeightGfpa; 

#include <algorithm>
using std::sort;

//*************************************************************************
//
//                            Out of class
//
//*************************************************************************

int CmpWeightfpa(const void * a, const void *b)
  //*******************************************
{
  if(WeightGfpa[*(unsigned int *)a] > WeightGfpa[*(unsigned int *)b]) return -1;
  if(WeightGfpa[*(unsigned int *)a] < WeightGfpa[*(unsigned int *)b]) return  1;
  return 0;
}
struct WeightfpaCmp
{
  bool operator()(const unsigned& a, const unsigned& b) const
  {
    return WeightGfpa[a]>WeightGfpa[b];
  }
};

double FpaCrisp(int nb, double *mu, double *Tab)
  //********************************************
{
  double sum = 0.0; 
  double wg = 0.0;
  int i;

  if(!nb) return FisMknan();
 
  for(i = 0; i < nb; i ++)
    {
      sum += mu[i] * Tab[i];
      wg += mu[i];
    }
  return sum / wg;
}

int FpaFuzzy(int nb, double *mu, double *Tab, FISOUT *S) 
  //****************************************************
{
  int i, nmf, max_mf;
  double crisp, max;
  
  nmf = S->GetNbMf();
  if(nmf < 2)  return 0;

  crisp = FpaCrisp(nb, mu, Tab);
  if(FisIsnan(crisp)) return 0;

  S->GetDegsV(crisp);
  max = S->Mfdeg()[0];
  max_mf = 0;

  for(i = 1; i < nmf; i++)
    {
      if(S->Mfdeg()[i] > max)
	{
	  max = S->Mfdeg()[i];
	  max_mf = i;
	}
    }
  return max_mf + 1;
}

double FpaClassif(int nb, double * Tab) 
  //***********************************
{
  if( ! nb) return -1;

  double * val, * Unique, ret; 
  int * nu, nbUnique, max, i_max, i;

  val = new double [nb];

  for(i = 0; i < nb; i ++)  val[i] = Tab[i];

  SortUniq(val, nb, &Unique, &nu, nbUnique, 0.01);

  max = nu[0];
  i_max = 0;
  for(i = 1; i < nbUnique; i ++)
    if(nu[i] > max) 
      {
	max = nu[i];
	i_max = i;
      }

  ret = Unique[i_max];

  delete [] val;
  delete [] Unique; delete [] nu;
  return ret;
}



//*************************************************************************
//
//                            Class FISFPA
//
//*************************************************************************

int FISFPA::FpaRules(int nout) 
  //**************************
{
  int i;
  int *sorted;
  int ret;
  int card; 
  SortDeg * degs;

  if(NbIn + nout > NbCol) return -1;

  WeightGfpa = new double [NbEx];
  sorted = new int [NbEx];
  degs = new SortDeg [NbEx];

  ret = 0;

  for(int k = 0; k < NbIn; k++) In[k]->privMfdeg=true;

  for(i = 0; i < NbRules; i++)    
    {  
 #ifdef _OPENMP
      const std::exception *err=NULL;
#pragma omp parallel 
     {
      try { 
#pragma omp for schedule(dynamic)
#endif
      for(int j = 0; j < NbEx; j++) 
	{
	  for(int k = 0; k < NbIn; k++)  // Fuzzification
	    {
	      if(In[k]->IsActive() == false) continue;

	      if(FisIsnan(Tab[j][k]))            // missing value
		{
		  if(!strcmp(strMissingValues, FIS::RandomMissingValues()))
		    In[k]->GetRandDegs(Tab[j][k]);
		  else if(!strcmp(strMissingValues, FIS::MeanMissingValues()))
		    In[k]->SetEqDegs(Tab[j][k]); 
		  else 
		    {  
		      sprintf( ErrorMsg, "~UnknownMissingValueStrategy~: %.50s", strMissingValues );
		      throw std::runtime_error( ErrorMsg );
		    }
		}
	      else if(In[k]->GetDegs(Tab[j][k])) ret = k;

	    }
	  
	  // Choix d'optimisation espace au lieu de temps : 
	  // les degres d'appartenance sont recalcules pour chaque regle
	  // alternative : les stocker.
	  	  	  
	  WeightGfpa[j] = Rule[i]->MatchDeg();
	  sorted[j] = j;
	  
	} // end of for j
#ifdef _OPENMP
      }	
// catch exception, and pass out of parallel region
      catch (const std::exception& ex)
        {
          err=&ex;
        }
      }
      if (err) throw err;
#endif
      // Sort the index array according to the matching degree values
      //qsort(sorted, NbEx, sizeof(unsigned int), CmpWeightfpa);
     
      sort(sorted,sorted+NbEx,WeightfpaCmp());

      for(int j = 0; j < NbEx; j++)     // store in SortDeg struture
	{
	  degs[j].deg = WeightGfpa[sorted[j]];
	  degs[j].num = sorted[j];
	}

      // Calculation of the cardinality of the set that will be used 
      if(Strategy == MIN)       card = SelErMin(degs);
      else if(Strategy == DEC) 	card = SelErDec(degs);
      else 
	{
	  sprintf( ErrorMsg, "~UnknownCardinalityCalculStrategy~: %.50s", strMissingValues );
	  throw std::runtime_error( ErrorMsg );
	}

      if(Out[nout]->Classification() && 
	 ! strcmp(Out[nout]->GetOutputType(), OUT_CRISP::OutputType()))     
	FisfpaClassif(i, card, degs, nout);
      else if(! strcmp(Out[nout]->GetOutputType(), OUT_CRISP::OutputType())  )  
	FisfpaCrisp(i, card, degs, nout);
      else   FisfpaFuzzy(i, card, degs, nout);
    } // end of for i

  for(int k = 0; k < NbIn; k++) In[k]->privMfdeg=false;

  delete [] WeightGfpa;
  delete [] sorted;
  delete [] degs;

  
  return ret;
} // End of FpaRules()


double FISFPA::FpaARule(int nrule, int nout, int nochange) 
  //******************************************************
{
  int j, k;
  int *sorted;
  int ret;
  int card; 
  double indexH, rweight;
  SortDeg * degs;

  if(NbIn + nout > NbCol) return -1.;

  WeightGfpa = new double [NbEx];
  sorted = new int [NbEx];
  degs = new SortDeg [NbEx];
  ret = 0;

#ifdef _OPENMP
  for(int k = 0; k < NbIn; k++) In[k]->privMfdeg=true;
  const std::exception *err=NULL;
#pragma omp parallel 
  {
    try 
      { 
#pragma omp for schedule(dynamic)
#endif
  for(j = 0; j < NbEx; j++) // Calculate the matching degree for each example
    {
      for(k = 0; k < NbIn; k++)  // Fuzzification
	{
	  if(In[k]->IsActive() == false) continue;
	  
	  if(FisIsnan(Tab[j][k]))            // missing value
	    {
	      if(!strcmp(strMissingValues, "aleatoire"))
		In[k]->GetRandDegs(Tab[j][k]);
	      else if(!strcmp(strMissingValues, "mean"))
		In[k]->SetEqDegs(Tab[j][k]); 
	      else 
		{  
		  sprintf( ErrorMsg, "~UnknownMissingValueStrategy~: %.50s", strMissingValues );
		  throw std::runtime_error( ErrorMsg );
		}
	    }
	  else if(In[k]->GetDegs(Tab[j][k])) ret = k;
	  
	}
      
      // Choix d'optimisation espace au lieu de temps : 
      // les degres d'appartenance sont recalcules pour chaque regle
      // alternative : les stocker.
      
      WeightGfpa[j] = Rule[nrule]->MatchDeg();
      sorted[j] = j;
      
    } // end of for j
      	
#ifdef _OPENMP
      }
    // catch exception, and pass out of parallel region
    catch (const std::exception& ex)
      {
        err=&ex;
      }
  }
  for(int k = 0; k < NbIn; k++) In[k]->privMfdeg=false;
  if (err) throw err;
#endif
      	
      // Sort the index array according to the matching degree values
  //qsort(sorted, NbEx, sizeof(unsigned int), CmpWeightfpa);
  sort(sorted, sorted+NbEx, WeightfpaCmp());
  
  for(j = 0; j < NbEx; j++)     // store in SortDeg struture
    {
      degs[j].deg = WeightGfpa[sorted[j]];
      degs[j].num = sorted[j];
    }
  
   if(nochange == false)
    {
      // Calculation of the cardinality of the set that will be used 
      if(Strategy == MIN) card = SelErMin(degs);
      else if(Strategy == DEC) 	card = SelErDec(degs);
      else 
	{
	  sprintf( ErrorMsg, "~UnknownCardinalityCalculStrategy~: %.50s", strMissingValues );
	  throw std::runtime_error( ErrorMsg );
	}

      if(card < MinCard) return NOT_ENOUGH;

      if(Out[nout]->Classification() &&     
	 ! strcmp(Out[nout]->GetOutputType(), OUT_CRISP::OutputType()) )
	FisfpaClassif(nrule, card, degs, nout);
      else if(! strcmp(Out[nout]->GetOutputType(), OUT_CRISP::OutputType())   )  
	FisfpaCrisp(nrule, card, degs, nout);
      else   FisfpaFuzzy(nrule, card, degs, nout);
    }

  card = SelErMin(degs);
  if(card < MinCard) return NOT_ENOUGH;

  if( ! strcmp(Out[nout]->GetOutputType(), OUT_CRISP::OutputType()) && Out[nout]->Classification())
    {
      int ncl =0;
      double * lbs=NULL;

      if (! strcmp(Out[nout]->Defuzzify(), OUT_CRISP::SugenoDefuz()))
	{
	  DEFUZ_SugenoClassif * D = (DEFUZ_SugenoClassif *) Out[nout]->Def;
	  ncl = D->NbClasses();
	  lbs = D->ClasLabel();
	}
      else if (! strcmp(Out[nout]->Defuzzify(), OUT_CRISP::MaxCrispDefuz()))
	{
	  DEFUZ_MaxCrisp * D = (DEFUZ_MaxCrisp *) Out[nout]->Def;
	  ncl = D->NbClasses();
	  lbs = D->ClasLabel();
	}
	
      else
	{
	  sprintf( ErrorMsg, "~UnknownDefuzzification~: %.100s~", Out[nout]->Defuzzify());
	  throw std::runtime_error( ErrorMsg );
	}
      indexH = ComputeIndexClassif(card, degs, nout, rweight, ncl, lbs);
    } 

  else indexH = ComputeIndexReg(card, degs, nout, rweight);

  Rule[nrule]->Weight = rweight;

  delete [] WeightGfpa;
  delete [] sorted;
  delete [] degs;
  
  return indexH;
} // End of FpaARule()



int FISFPA::SelErDec(SortDeg * v) 
  //*****************************
{
  int i;
  double degre;
  int card = 0;

  for(degre = START_DEC; degre > MinDeg; degre -= STEP_DEC)
    {
      for(i = 0; i < NbEx; i++)  if(v[i].deg < degre - PRECISION_DEGRE) break;
      card = i;
      if(card >= MinCard) break;
      card = 0; // out with a card = 0
    }
  return card;
}


int FISFPA::SelErMin(SortDeg * v) 
  //*****************************
{
  int i;

  for(i = 0; i < NbEx; i++)
    if(v[i].deg < MinDeg - PRECISION_DEGRE) break;
  
  if(i < MinCard) return 0;
  
  return i;
}

void FISFPA::FisfpaCrisp(int nrule, int nb, SortDeg * v, int nout) 
  //**************************************************************
{
  if( ! nb)
    {
      Rule[nrule]->SetAConc(nout, Out[nout]->DefaultValue());
      Rule[nrule]->Deactivate();
      return;
    }

  double *T, *deg, value;
  int i;

  T = new double [nb];
  deg = new double [nb];

  for(i = 0; i < nb; i ++) 
    {
      T[i] = Tab[v[i].num][NbIn + nout];
      deg[i] = v[i].deg;
    }

  value = FpaCrisp(nb, deg, T);
  delete [] T; delete [] deg;

  Rule[nrule]->SetAConc(nout, value);
}


void FISFPA::FisfpaFuzzy(int nrule,  int nb, SortDeg * v, int nout) 
  //***************************************************************
{
  if( ! nb)
    {
      // There is at least 1 MF in a fuzzy output 
      Rule[nrule]->SetAConc(nout, 1); 
      Rule[nrule]->Deactivate();
      return;
    }

  double *T, *deg;
  int i;

  T = new double [nb];
  deg = new double [nb];

  for(i = 0; i < nb; i ++) 
    {
      T[i] = Tab[v[i].num][NbIn + nout];
      deg[i] = v[i].deg;
    }

  i = FpaFuzzy(nb, deg, T, Out[nout]);
  delete [] T; delete [] deg;

  Rule[nrule]->SetAConc(nout, i);
}


void FISFPA::FisfpaClassif(int nrule, int nb, SortDeg * v, int nout) 
  //****************************************************************
{
  if( ! nb)
    {
      Rule[nrule]->SetAConc(nout, Out[nout]->DefaultValue());
      Rule[nrule]->Deactivate();
      return;
    }

  double * T, valc;
  T = new double [nb];
  int i;
  for(i = 0; i < nb; i ++) T[i] = Tab[v[i].num][NbIn + nout];

  valc = FpaClassif(nb, T);
  delete [] T;
 
  Rule[nrule]->SetAConc(nout, valc);
}

double FISFPA::ComputeIndexClassif(int nb, SortDeg * v, int Nout, double & w, int nclasses, const double * labels)
  //**************************************************************************************************************
{
  //entropy based heterogeneity index (Pielou index)
  // nb is the number of examples
  int i, k, l, numcase;
  double * card, index, sum, pk;
  
  // protection
  if (nclasses < 2)
    {
      sprintf(ErrorMsg, "~NoClassFoundInFISFPAmclasses=::ComputeIndexClassif~:%d\n", nclasses);
      throw std::runtime_error( ErrorMsg );
    }
    
  card = new double [nclasses];
  for(i = 0; i < nclasses; i++) card[i] = 0.;
  w = 0;
  
  for(i = 0; i < nb; i++) 
    {
      w += v[i].deg;  
      numcase = -1;
      for(l = 0; l < nclasses; l++)
	if(fabs(Tab[v[i].num][NbIn + Nout] - labels[l]) < EPSILON2) numcase = l;
      
      if(numcase == -1) 
	{
	  sprintf(ErrorMsg, "~LabelNotFoundInFISFPA::ComputeIndexClassif~:%f\n",Tab[v[i].num][NbIn + Nout]);
	  throw std::runtime_error( ErrorMsg );
	}
      card[numcase] += v[i].deg;
    }

  sum = 0.;
  for(i = 0; i < nclasses; i++)  sum += card[i];
  // The sum could be computed in the last cell of the array
  if(sum < EPSILON) index = -INFINI;
  else 
    {
      index = 0.0;
      for(k = 0; k < nclasses; k++)
	{
	  if(card[k] > EPSILON)
	    {
	      pk = card[k] / sum;
	      index -= pk * log(pk);
	    }
	}
    }
  index = index / log((double)nclasses);
  
  delete [] card;

  return index;
}

double FISFPA::ComputeIndexReg(int nb, SortDeg * v, int Nout, double & w)
  //*********************************************************************
{
  int i;
  double index, sum, mean, tmp;

  sum = tmp = 0.0;
  for(i = 0; i < nb; i++) 
    {
      sum += v[i].deg;  
      tmp +=  Tab[v[i].num][NbIn + Nout] * v[i].deg;
    }
  mean = tmp / sum;
  
  tmp = 0.0;
  for(i = 0; i < nb; i++) 
    tmp +=  (Tab[v[i].num][NbIn + Nout]  - mean) *  (Tab[v[i].num][NbIn + Nout]  - mean) * v[i].deg;
  
  index = sqrt(tmp / sum);
  w = sum;

  return index/stdv[Nout];
}

//****************    End of  FPA.CPP   ***********************
