//***********************************************************************
//
//
//                              FPA.H
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : FISFPA class functions used by FISPRO, 
//        part of package fpa (Fast Prototype Algorithm)
//        requires library fispro

//**********************************************************************
#ifndef __FPA_H
#define __FPA_H


#include "fis.h"

//! Startegy for computing the cardinality of the set used to compute the rule conclusion (see user documentation for details)
#define MIN 0
//! Startegy for computing the cardinality of the set used to compute the rule conclusion  (see user documentation for details)
#define DEC 1
//! Internal parameter for the 'DEC' startegy 
#define START_DEC 0.7 
//! Internal parameter for the 'DEC' startegy 
#define STEP_DEC  0.1 
//! Return value used by FpaARule if the number of items which fire the rule 
//! is less than the parameter 'MinCard'
#define NOT_ENOUGH  1.0011 
//! Tolerance value for membership degree comparison
#define PRECISION_DEGRE 0.01

#ifndef __SORTDEG
#define __SORTDEG
struct SortDeg
{
  //! Matching degree
  double deg;   
  //! Example number
  int num;      
};
#endif


//! Function called by qsort to sort the SortDeg structure
int CmpWeightfpa(const void * a, const void *b);

//! Initialize a crisp rule conclusion using a weighted sum of observed values. 
//! 'Tab' contains the output values sorted according the item degree of match, 
//! 'mu' the corresponding matching degrees, only the  'nb' first are used. 
//! Return:  the rule conclusion or NaN if  'nb' is 0. 
double FpaCrisp(int nb, double *mu, double *Tab);

//! Iinitialize a fuzzy rule conclusion. 
//! 'Tab' contains the output values sorted according the item degree of match, 
//! 'mu' the corresponding matching degrees, only the  'nb' first are used. 
//! 'S' is the output pointer. 
//! Return the MF number of 'S' or NaN if 'nb' is 0. 
//! (The first MF is is the number 1, not 0). 
int FpaFuzzy(int nb, double *mu, double *Tab, FISOUT *S); 

//! Initialize a crisp conclusion using a majority vote. 
double FpaClassif(int nb, double *Tab); 



//! The inherited class from FIS to initialize rule conclusions.
class FISFPA : public FIS
//************************
{
  protected :

  //! The number of examples in the training set
  int NbEx;     
  //! The number of variables (columns) in the training set 
  int NbCol;
  //! The sample data array    
  double ** Tab;      

  //! The standard deviation array (dim = number of outputs)
  double * stdv;
  
  //! The minimum cardinality to compute the conclusion
  int MinCard;     
  //! The minimum matching degree of the example to be part of the set 
  //! used for computing the conclusion 
  double MinDeg;     
  //! Two possibilities MIN or DEC (0 ou 1). See user documentation for detail.                                      
  int Strategy;       
     
  //! Item selection according to the MIN startegy:  
  //! All the items whose degree is greater than 'MinDeg' are in the set. 
  //!  Return value : the cardinality of the set.
  int SelErMin(SortDeg * v);

  //! Item selection according to the DEC startegy:  
  //! Only the items who fire the most the rule are part of the set. 
  //! The threshold is first set to START_DEC and while card is less than 'MinCard' 
  //! it is decreased of STEP_DEC until it reaches 'MinDeg'. 
  //! Return value : the cardinality of the set.
  int SelErDec(SortDeg * v);

  //! Call FpaCrisp to initialize the conclusion of rule number 'nrule'. 
  //! 'nout' is the output number, 'nb' is the number of samples to take into account, 
  //! 'v' is the SortDeg structure. 
  //! If 'nb' is zero, the initialization is done with the default output value.
  void FisfpaCrisp(int nrule, int nb, SortDeg * v, int nout);

  //! Call FpaClassif to initialize the conclusion of rule number 'nrule'. 
  //! 'nout' is the output number, 'nb' is the number of samples to take into account,  
  //! 'v' is the SortDeg structure. 
  //! If 'nb' is zero, the initialization is done with the default output value.
  void FisfpaClassif(int nrule, int nb, SortDeg * v, int nout);

  //! Call FpaFuzzy to initialize the conclusion of rule number 'nrule'. 
  //! 'nout' is the output number, 'nb' is the number of samples to take into account,  
  //! 'v' is the SortDeg structure. 
  //! If 'nb' is zero, the initialization is done with the default output value.
  void FisfpaFuzzy(int nrule, int nb, SortDeg * v, int nout);



  //! Return: entropy based entropy based heterogeneity index, standardized with log(nclasses)
  //! 'w' is the cumulated weight of the first 'nb' items.
  double ComputeIndexClassif(int nb, SortDeg * v, int Nout, double & w, int nclasses, const double * labels);
  //! Return : standard deviation within the first 'nb' items. 
  //! 'w' is the cumulated weight of the first 'nb' items.
  double ComputeIndexReg(int nb, SortDeg * v, int Nout, double & w);

  void operator=(const FISFPA& x);
  FISFPA(const FISFPA&);
  public :

    //! Initialize the rule conclusions of all the rules according to the 
    //! parameters of class FISFPA. 'nout' is the number of the considered output. 
    //! Return value :  0 if OK, +x if all the degrees are zero for input x, 
    //! -1 if observed data are not available for output 'nout'
    int FpaRules(int nout) ;

  //! Initialize the conclusion of the rule number 'nrule' according to the 
  //! parameters of class FISFPA. 'nout' is the number of the considered output. 
  //! When 'nochange' is true rule conclusion is not changed. 
  //! Return value :  output homogeneity index, NOT_ENOUGH if card < 'MinCard', 
  //! -1 if observed data are not available for output 'nout.
  double FpaARule(int nrule, int nout, int nochange = false); 

  virtual void Print(FILE * f = stdout) const
    {
      fprintf(f, "\n Strategy : %d  MinCard : %d  MinDeg : %4.2f", Strategy, MinCard, MinDeg);
      FIS::Print(f);
    }

  FISFPA(const FIS& fis,const char * Examples, int strat, int effmin, double degmin):
    FIS(fis), Tab(NULL), stdv(NULL) {set(Examples,strat,effmin,degmin);}

  FISFPA(const char * Config, const char * Examples, int strat, int effmin, double degmin) : FIS(Config), Tab(NULL), stdv(NULL) {set(Examples,strat,effmin,degmin);}
    
  void set(const char * Examples, int strat, int effmin, double degmin)
    {
      int i, j;
      double *tmp, median, mean, min, max;
      delete [] Tab;
      delete [] stdv;

      NbCol = NbIn + NbOut;
      Tab = ReadSampleFile(Examples, NbCol, NbEx);
      stdv = new double [NbOut];
      tmp = new double [NbEx];

      for(i = 0; i < NbOut && NbIn + i < NbCol; i++)
	{
	  stdv[i] = -1;
	  for(j = 0; j < NbEx; j++) tmp[j] = Tab[j][NbIn + i];

	  if(! strcmp(Out[i]->Defuzzify(), OUT_CRISP::MaxCrispDefuz()) )
	    ((DEFUZ_MaxCrisp *) Out[i]->Def)->InitClasses(tmp, NbEx);
	  
	  else if(Out[i]->Classification() && 
		  (! strcmp(Out[i]->GetOutputType(), OUT_CRISP::OutputType())) && 
		  (! strcmp(Out[i]->Defuzzify(), OUT_CRISP::SugenoDefuz())) )
	    ((DEFUZ_SugenoClassif *) Out[i]->Def)->InitClasses(tmp, NbEx);

	  else
	    StatArray(tmp, NbEx, 0, median, mean, stdv[i], min, max); 
	}
      delete [] tmp;
      Strategy = strat;
      MinCard = effmin;
      MinDeg = degmin; 
    }
  

  

  ~FISFPA()
    {
      for(int j = 0; j < NbEx; j++) delete [] Tab[j];
      delete [] Tab;
      delete [] stdv;
    }
};
#endif
//****************    End of  FPA.H   ***********************
