//****************************************************************
//
//
//                              GENFIS.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Functions in classes derived from RULE and FIS to generate
//         the rule  premise from input partitioning
//        and optionally from a data file

// Environnment : requires library fispro

//****************************************************************



#include "genfis.h"

//****************************************************************
//
//
//             Global function and variable used by qsort
//
//
//****************************************************************

//! Global variable used by qsort
double * CumulG; 

//! Global function used by qsort
int CmpCummul(const void * a, const void *b)
  //****************************************
{
  if(CumulG[*(unsigned int *)a] > CumulG[*(unsigned int *)b]) return -1;
  if(CumulG[*(unsigned int *)a] < CumulG[*(unsigned int *)b]) return  1;
  return 0;
}


//! Init static variables 
int GENRULE::nI = 0, GENRULE::nO = 0;
FISIN **GENRULE::E = NULL;
char *GENRULE::conj = NULL;


//****************************************************************
//
//
//                Class GENFIS
//
//
//****************************************************************

void GENFIS::InitSystem(char * fic, const char * dataf)
  //*********************************************
{
  int i, bsize;

  // Remove existing rules and MFConcs
  for (int i=0;i<NbOut;i++)
    {
      DeleteMFConc(i);
      DeleteMFConcArray(i);
    }
  
  if (Rule!=NULL)
    {
      for(i = 0; i < NbRules; i++) delete Rule[i];
      delete [] Rule;
    }
  NbRules = 0;
  Rule = NULL;
  
  
  if(dataf) GenereRules(dataf);
  
  else GenereRules();  // All possible combinations
  
  NbActRules = NbRules;
  //init conclusions for printing FIS configuration files
  for(i = 0; i < NbRules; i++)
    {
      Rule[i]->SetConclusion(NbOut, Out);
      for (int iout=0;iout<NbOut;iout++)
	Rule[i]->SetAConc(iout,1.0);
    }

  ifstream f(fic);
  if (f.fail())
    {
      sprintf( ErrorMsg, "~CannotOpenFISFile~: %.100s~", fic);
      throw std::runtime_error( ErrorMsg );
    }
  bsize = MaxLineSize(f);
  if(NbExceptions) ReadExcep(f, bsize);
}


void GENFIS::ReadExcep(ifstream &f, int bufsize)
  //********************************************
{
  RULE * TmpRule;
  int i, pos, from;
  char *tmp, *buf;

  tmp = new char[bufsize];
  buf = new char[bufsize];

  sprintf(tmp, "[Exceptions]");
  pos = strlen(tmp);
  
  do{
    f.getline(buf, bufsize);
  }while(strncmp(tmp, buf, pos));
  // The tag "[Exceptions]" exists:  it has been read by FIS::InitSystem

  for(i = 0; i < NbExceptions; i++)
    {
      f.getline(buf, bufsize);
      TmpRule = new RULE( NbIn, In, 0, NULL, cConjunction, buf);
       
      pos = from = 0;
       
      while( (pos = RulePos(TmpRule, from)) != -1)
	{
	  Rule[pos]->Deactivate();
	  from = pos + 1;
	  NbActRules --;
	};
      delete TmpRule;
    } 
   
  delete [] tmp;
  delete [] buf;
}



void GENFIS::Print(FILE *f) const
  //***********************
{
  int i;
  
  fprintf(f,"\nSystem : %s", Name);
  fprintf(f,"\nNumber of inputs : %d\tNumber of outputs : %d\n", NbIn, NbOut);
  
  fprintf(f,"\nNumber of rules : %d\tNumber of exceptions : %d", NbRules, NbExceptions) ;

  fprintf(f,"\nConjunction : %s", cConjunction);
  fprintf(f,"\nMissing value management : %s\n", strMissingValues);
    
  for(i = 0; i < NbIn; i++)   In[i]->Print(f);
  for(i = 0; i < NbOut; i++)  Out[i]->Print(f);

  fprintf(f,"\nRules (the weight is in the last column) : \n");
     
  for(i = 0; i < NbRules; i++)   
    if(Rule[sorted[i]]->IsActive()) Rule[sorted[i]]->Print(f);
}

void GENFIS::PrintCfg(FILE *f, const char *fd) const
  //************************************************
{
  int i, n;

  for(i = 0; (unsigned) i < strlen(Name); i++)    if(isalnum(Name[i])) break;
  if((unsigned) i == strlen(Name)) sprintf(Name, "system");  
  // None alphanumeric character in the string Name


  n = 0;
  for(i = 0; i < NbRules; i++)
      if(Rule[sorted[i]]->IsActive() && Rule[sorted[i]]->Weight > CumWeThres) n++;
 
  fprintf(f, "[System]\n");
  fprintf(f, "Name=%c%s%c\n", STRING_SEP, Name, STRING_SEP);
  fprintf(f, "Ninputs=%d\n", NbIn);
  fprintf(f, "Noutputs=%d\n", NbOut);
  fprintf(f, "Nrules=%d\n", n);
  fprintf(f, "Nexceptions=0\n");
  fprintf(f, "Conjunction=%c%s%c\n", STRING_SEP, cConjunction, STRING_SEP);
  fprintf(f, "MissingValues=%c%s%c\n", STRING_SEP, strMissingValues, STRING_SEP);

  for(i = 0; i < NbIn; i++)   In[i]->PrintCfg(i + 1, f, fd);

  for(i = 0; i < NbOut; i++)  Out[i]->PrintCfg(i + 1, f, fd);

  fprintf(f, "\n[Rules]\n");
  for(i = 0; i < NbRules; i++)   
    if(Rule[sorted[i]]->IsActive() && Rule[sorted[i]]->Weight > CumWeThres) 
      Rule[sorted[i]]->PrintCfg(f, fd);

  fprintf(f, "\n[Exceptions]\n");

} 

//****************************************************************
//
//
//                        Rule generation 
//
//
//****************************************************************

void GENFIS::GenereRules(void)
  //**************************
{ 
  int i;
  unsigned int j, Nb, old;
  int *active;

  active = new int[NbIn];

  Nb = old = 1;
  
  for(i = 0; i < NbIn; i++) 
    {
      if(!  In[i]->GetNbMf()) In[i]->Deactivate();;
      active[i] = In[i]->IsActive();
      if(active[i]) Nb *= In[i]->GetNbMf();  // Nb contains the max number of rules
      if(Nb < old) 
	{
	  sprintf( ErrorMsg, "~TooManyRulesToGenerate~\n~NumberOfRules~ > %u\n", Nb);
	  throw std::runtime_error( ErrorMsg );
	}
      else old = Nb;
    }

  if(Nb > MAX_RULES)
    {
      sprintf( ErrorMsg, "~TooManyRulesToGenerate~\n~NumberOfRules~ = %u (~MaxNumber~= %u) \n", 
	       Nb, MAX_RULES);
      throw std::runtime_error( ErrorMsg );
    }

  CumulG = NULL;
  Rule = NULL;
  GENRULE::nI = NbIn;
  GENRULE::nO = NbOut;
  GENRULE::E = In;
  GENRULE::conj = cConjunction;
  try
    {
      CumulG = new double [Nb];
      Rule = new RULE * [Nb];
      for(j = 0; j < Nb; j++)  Rule[j] = NULL;
    }

  catch( std::exception &e )
    {
      if(CumulG) delete [] CumulG;
      if(Rule) delete [] Rule;
      sprintf( ErrorMsg, "~TooManyRulesToGenerate~\n~NumberOfRules~ = %u\n", Nb);
      throw std::runtime_error( ErrorMsg );
    }

  Temp = new GENRULE;
  ValProps = new int [NbIn];
  CurrentCfg = new int [NbIn];
  for(i = 0; i < NbIn; i++)    CurrentCfg[i] = 0; // ANY
 
  NbRules = 0;     // Number of generated rules 

  GenereCont(0);  // without sample file, all possible combinations
  
  sorted = new unsigned int [NbRules];
  
  for(i = 0; i < NbRules; i++)     
    {
      sorted[i] = i;
      Rule[i]->Weight = 1. + CumWeThres;
    }

  delete [] CumulG;
  delete [] active;
} // End of GenereRules(istream *f)


void GENFIS::GenereCont(int i, int flag)
  //************************************
{
  int m, skip;
  
  if(i == NbIn - 1)    // The last input
    {  
      if(! In[i]->IsActive()) 	GenereAddRule(flag); 
         
      else
	{
	  for(m = 0; m < In[i]->GetNbMf(); m++)
	    {
	      skip = 0;
	      CurrentCfg[i] = m;
	      
	      if(flag && (! In[i]->Mfdeg()[m])) skip = 1; 
	      if ( skip == 0) GenereAddRule(flag);
	    } 
	} 
    } 
  else
    {
      if(! In[i]->IsActive()) GenereCont(i+1, flag); // using  CurrentCfg[i] = 0
      
      else
	{
	  for(m = 0; m < In[i]->GetNbMf(); m++)
	    {
	      CurrentCfg[i] = m;	  
	      if(flag && (! In[i]->Mfdeg()[m])) continue; 
	      GenereCont(i+1, flag);
	    }
	}
    }
} // End of GenereCont()


void GENFIS::GenereAddRule(int fl)
  //******************************
{
  int j, add, index;


  for(j = 0; j < NbIn; j++)
    {
      if(In[j]->IsActive()) ValProps[j] = CurrentCfg[j] +1;
      else ValProps[j] = CurrentCfg[j]; // ANY
    }
  
  Temp->SetAProps(ValProps);
  
  add = 1;
  
  if(fl)
    {
      Temp->ExecRule();
      index = RulePos(Temp);
      if(index != -1)  // already stored
	{
	  add = 0;
	  CumulG[index] += Temp->Weight;
	}
    }
  if(add)  // new rule
    {
      Rule[NbRules]= new RULE(*Temp, Temp->E);
      CumulG[NbRules] = Temp->Weight; 
      NbRules++;
    }
} // End of GenereAddRule


void GENFIS::GenereRules(const char * fdata) 
  //**********************************
{ 
  double ** Data, CumulW;
  int i, j, k, l, m, n, mf;
  int *active;
  int NbSave, NbCurrent;
  GENRULE *tmp;
  GENRULE ** Save;
  GENRULE ** Current;

  active = new int[NbIn];
  Data = NULL;

  CumulG = NULL;
  Current = NULL;
  NbCurrent = 0;
  CumulW = 0.0;

  GENRULE::nI = NbIn;
  GENRULE::nO = NbOut;
  GENRULE::E = In;
  GENRULE::conj = cConjunction;

  Save = new GENRULE * [1]; 
  NbSave = 1;
  Save[0] = new GENRULE();

  Data = ReadSampleFile(fdata, j, n);

  for(i = 0; i < NbIn; i++) 
    {
      if(In[i]->GetNbMf() < 2) In[i]->Deactivate();
      active[i] = In[i]->IsActive();
    }

  for(i = 0; i < NbIn; i++) 
    {
      if(! In[i]->IsActive()) continue;
      if(Current) 
	{
	  for(j = 0; j < NbCurrent; j++)
	    if(Current[j])  delete Current[j];
	  delete [] Current; 
	}
      mf = In[i]->GetNbMf();
      Current = new GENRULE * [NbSave * mf];
      NbCurrent = 0;
      for(m = i+1; m < NbIn; m++) In[m]->Deactivate();  
      for(j = 0; j < NbSave; j++)
	{
	  for(k = 0; k < mf; k++)
	    {
	      tmp = new GENRULE(Save[j], In);
	      tmp->Weight = 0.0;   // The copy constructor also save the Weight.
	      tmp->SetAProp(k+1, i);
	      CumulW = 0.0;
	      for(l = 0; l < n; l++)
		{
		  for(m = 0; m < NbIn; m++)      // Get mf degrees
		    {
		      if(! In[m]->IsActive() ) continue;
		      if(FisIsnan(Data[l][m]))    // missing value 
			In[m]->SetEqDegs(Data[l][m]);
		      else In[m]->GetDegsV(Data[l][m]); 
		    }
		  tmp->ExecRule();
		  CumulW += tmp->Weight;
		  if((i < (NbIn-1)) && (CumulW >= CumWeThres)) break;
		}
	      if(CumulW >= CumWeThres) 
		{
		  Current[NbCurrent] = new GENRULE(tmp, In);
		  Current[NbCurrent]->Weight = CumulW;
		  NbCurrent ++;
		}
	      delete tmp;
	    }
	}
      for(m = i+1; m < NbIn; m++) if (active[m]) In[m]->Activate();  
      
      for(j = 0; j < NbSave; j++) if(Save[j]) delete Save[j];
      delete [] Save;
      Save = new GENRULE * [NbCurrent];
      for(j = 0; j < NbCurrent; j++) 
	  Save[j] = new GENRULE(Current[j], In);
	
      NbSave = NbCurrent;
   }

  for(j = 0; j < NbSave; j++) if(Save[j]) delete Save[j];
  delete [] Save;
  Save = NULL;

  CumulG = new double [NbCurrent];

  sorted = new unsigned int [NbCurrent];
  for(i = 0; i < NbCurrent; i++)     
    {
      sorted[i] = i;
      CumulG[i] = Current[i]->Weight;
    }

  if(SortCumulg)
      qsort(sorted, NbCurrent, sizeof(unsigned int), CmpCummul);

  // Detruire Current, remplir Rule
  NbRules = NbCurrent;

  // delete existing rules
  if (Rule !=NULL) 
    {
      for(i = 0; i < NbRules; i++) //check Nbrule !!!
	delete [] Rule;
      Rule=NULL;
    }
  
  
  Rule = new RULE * [NbRules];
  for(j = 0; j < NbCurrent; j++)
    Rule[j] = new RULE(*(Current[j]), In);

  if (Data !=NULL)
    {
      for(i = 0; i < n; i++) 
	delete [] Data[i]; 
      delete [] Data;
      Data=NULL;
    }
  if (Current !=NULL)
    {
      for(i = 0; i < NbCurrent; i++) 
	delete Current[i];   
      delete [] Current;
      Current=NULL;
    }
  delete [] CumulG;
  CumulG=NULL;
  delete [] active;
  active=NULL;
} // End of GenereRules(istream *f)

//***********************  End of GENFIS.CPP    ***************
