
//****************************************************************
//
//
//                              GENFIS.H
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : Functions in classes derived from RULE and FIS to generate
//         the rule  premise from input partitioning
//        and optionally from a data file

// Environnment : requires library fispro
//****************************************************************

#ifndef __GENFIS_H
#define __GENFIS_H

#include "fis.h"

//! This number is 1024*1024, it also corresponds to 10 variables with 4 MF.
#define MAX_RULES 1048576

//****************************************************************
//
//
//                        CLASS GENRULE
//
//
//****************************************************************

//! This class is used for rule generation. It does not manage the conclusion 
//! part of the rule.

class GENRULE : public RULE
//*************************
{
  protected :

    //! The number of outputs in the FIS. Their values will be set to 1.0 in each rule.
    int miss;     
  
  public :

    static int nI, nO;
    static FISIN **E;
    static char *conj;

    void Print(FILE * f = stdout) const
    {
      int i;
      double MissVal = 1.0;
      
      Prem->Print(f);
      for(i = 0; i < miss; i++) fprintf(f, "%8.3f%c ", MissVal, SEPARE);
      fprintf(f, "%8.3f%c ", Weight, SEPARE);
      fprintf(f, "\n");
    }

  void PrintCfg(FILE * f = stdout, char *fd = FORMAT_DOUBLE) const
    {
      int i;
      double MissVal = 1.0;

      Prem->Print(f);
      for(i = 0; i < miss; i++) fprintf(f, "%8.3f%c ", MissVal, SEPARE);
      fprintf(f, "\n");
    }

  void SetAProps(int * Tab)
    {
      int i;
      int n = GetNbProp();
      int *V;

      V = new int [n];

      for(i = 0; i < n; i++) V[i]= Tab[i]; 
  
      Prem->SetAProps(V);

      delete [] V;
    }

  GENRULE() : RULE()
    { 
      SetPremise(nI, E, conj);

      Weight = 0.0;
      miss = nO; // for the Print function.
      Active = true;
    }

  GENRULE(GENRULE &r, FISIN ** E)
    {
      Init();
      Active = r.Active;
      Prem = r.Prem->Clone(E);
      miss = nO;
      Weight = r.Weight;
    }

  GENRULE(GENRULE *r, FISIN ** E)
    {
      Init();
      Active = r->Active;
      Prem = r->Prem->Clone(E);
      miss = nO;
      Weight = r->Weight;
    }
  
  
};


//****************************************************************
//
//
//                        CLASS GENFIS
//
//
//****************************************************************

//! Inherited class of FIS for rule generation from input partitioning ond, 
//! as an option, considering a data sample file.
class GENFIS : public FIS
//***********************
{
  GENFIS& operator=(const GENFIS&);
  GENFIS(const GENFIS&);
 protected:
  
  //! Rule corresponding to exceptions mentionned in the FIS configuration file 
  //! are deactivated.
  void ReadExcep(ifstream &f, int bufsize);

  //! 'fic' is the FIS configuration file, It is needed to manage the exceptions. 
  //! Rule corresponding to exceptions are not generated. 'fic' is passed to ReadExcep. 
  //! 'dataf' is the data set file name. 
  //! If the pointer 'dataf' is NULL all the rules corresponding to all 
  //! input combinations will be generated. Otherwise, only the ones fired by 
  //! an example will be stored. Call GenereRules with the file pointed by 'dataf'.
  void InitSystem(char *fic, const char * dataf) ;
 
  void GenereRules(void);   

  void GenereRules(const char * fdata);   

  //! Recursive fonction called by GenereRules. If 'flag' is true then the 
  //! sample file is taken into account.
  void GenereCont(int i, int flag = 0);  

  //! Fonction called by GenereCont to make rule instanciation. 
  //! If 'flag' is true then the sample file is used to compute the cumulated 
  //! weight of each rule over the examples.
  void GenereAddRule(int fl);

  //! The temporary rule under construction.
  GENRULE * Temp;
 
  //! The current configuration of the premise rule under construction. 
  //! Size of the array:  the number of inputs.
  int * CurrentCfg;

  //! Analog to 'CurrentCfg' but used to manage deactivated inputs.
  int * ValProps;
  
  //! To store the sorted rule, according their cumulated weight, indices.
  //unsigned int * sorted;

  //! The number of generated rules which remain active because they do not match 
  //! any exception. Only these rules will be stored.
  int NbActRules; 

  //! If true the generated rules are sorted according to their cumulated weight 
  //! over the data set. Otherwise, the lexicographic order is used.
  int SortCumulg; 

  //! When 'SortCumulg' is true, only rules whose cumulated weight is greater 
  //! than this threshold are kept (default value: 0.0).
  //double CumWeThres;

 public:

  unsigned int * sorted;
  double CumWeThres;

  GENFIS(char *fic, const char * p_data, int TrC, double thres = 0.0) : FIS(fic)
    {
      SortCumulg = TrC;
      CumWeThres = thres;
      sorted = NULL;
      CurrentCfg = NULL;
      ValProps = NULL;
      Temp = NULL;
      InitSystem(fic, p_data);
    }

  ~GENFIS()
    {
      delete [] sorted;
      delete [] CurrentCfg;
      delete [] ValProps;
      delete Temp;
    }
  //! Print only active rules.
  void Print(FILE *f = stdout) const ;

  //! Print a FIS configuration file with the active rules only. 
  //! The rules are printed in a separate file whose name is made up of the string
  //! 'Name' of the FIS configuration file header (or 'system' if this 
  //! string is empty) and the extension '.rules'.
  void PrintCfg(FILE *f = stdout, const char *fd = FORMAT_DOUBLE) const ;
}; 
#endif

//*********************      End of GENFIS.H      **********************
