//**********************************************************************
//
//
//                              HFPFIN.H
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : INHFP class description, uUsed with fispro, 
//        part of package hfp (Hierarchical fuzzy partitioning)
//        requires library fispro
//
//**********************************************************************
#ifndef __HFPIN_H
#define __HFPIN_H

#include "fis.h"

#define  INFINI_HFP 1e123

//! Used to initialize the MF numbers to manage the multiple belonging
#define INACTIVE -1 
     
//! Default value for the double equality threshold 
#define TOLERANCE EPSILON // 0.01   

//! Hierarchy types
#define T_HFP 1
#define S_HFP "hfp"
#define T_KMEANS 2
#define S_KMEANS "kmeans"
#define T_REG 3
#define S_REG "regular"
      

//! Structure to store fuzzy sets vertices. 
//! 'l' and 'r' are the left and right coordinates for a trapezoidal shape. 
//! 'p' stands for pure, is no more used
struct TVertex
{
  double l,r;   
  int p;        
};


class INHFP : public FISIN
//************************
{
  INHFP& operator=(const INHFP&);
  INHFP(const INHFP&);
 protected:
  //! The sample values. 
  double *Val; 
  //! The sample size
  int Size;  
  //! The unique values, the values of 'Val' whose difference greater than the equality 
  //! tolerance threshold. 
  double *Unique; 
  //! The occurrence number of eacuh of the unique values in 'Val'.
  int * Occur;    
  //! The number of unique values, size of the array 'Unique'.
  int NbUnique;  
  
  //! The type of hierarchy
  int HierType;

  //! The distances between the unique values. Sizes of the squared array: NuUnique 
  double ** Dist; 

  //! If true the distance function is the symbolic one, else the numerical one
  int Symbolic;  
  //! The number of MF in the initial partition (default 1)
  int NmfInit;   

  //! A pointer on vertex arrays. Each configuration is stored in a vertex array. 
  //! The number of vertices in the array of indice i is i+1. 
  //! The fisrt value of each array is the density standard deviation of the partition in field 'l', 
  //! and the output variance weighted by the MF mass in field 'r'. 
  TVertex ** Vertices;  
  int Nvertices;
  //! The maximum size of the partition (number of vertices stored)
  int NmfMax;

  //! Build the initial partition using unique values in the data distribution.
  //! 'equal' is the equality threshold to compare values. 
  //! Calls the function SortUniq and remove Nan unique values.
  void UniqInitPart(double equal); 

  //! Build the initial partition using the k-means algorith.
  //! 'ng' is the maximum number of groups (empty groups are removed)
  void KmeansInitPart(int ng);

  //! The internal distance:  'mu1' and 'mu2' are the membership degrees. 
  //! Function called by Distance and SimpliDist.
  inline double IntDist(double mu1, double mu2) {return fabs(mu1-mu2);}

  //! The external distance:  'a' and 'b' are the MF numbers.
  //! Function called by Distance and SimpliDist.
  double ExtDist(int a, int b);

  //! Compute the distances between all the pairs of unique values considering 
  //! their membership degrees to the MF numbers ranging between 'smin' 
  //! and 'smax'. Fill the Dist array.
  void Distance(int smin, int smax);

  //! Compute the fuzzy cardinality (or mass) of Mf number 'f' from the 
  //! unique value array.
  double EvalMass(int f); 

  //! Pointer initialization to NULL and size of arrays to 0. Default values. 
  void Init(void)
    {
      Nvertices = 0;
      NbUnique = 0;
      NmfInit = 0;

      Val = NULL;
      Unique = NULL;
      Occur = NULL;
      Dist = NULL;
      Vertices = NULL;

      Symbolic = 0;
      HierType = T_REG;
    }

  //! This function defenition depends on the MF shape. It is defined for 
  //! triangular shaped ones in the file hfpintri.cpp. The extremities of 
  //1 the ppartition are triangular shaped.
  void InitMfBordTri(double * v);

  //! This function definition depends on the MF shape. It is defined for 
  //! triangular shaped ones in the file hfpintri.cpp. The extremities of 
  //1 the ppartition are trapezoidal shaped.
  void InitMfBordTrap(double * v);

  //! Create the initial fuzzy partition, each of the unique values being 
  //! a fuzzy set center.
  //! This function definition depends on the MF shape. It is defined for 
  //! triangular shaped ones in the file hfpintri.cpp. 
  //! The extreme sets are trapezoidal shaped, call of InitMfBordTrap. 
  //! They can also be defined as triangular ones using the call of the 
  //! InitMfBordTri function.
  // Allocate a Nmf size fuzzy partition and fill it with the array passed 
  // as an argument
  void InitMf(double * v)
    {
      InitMfBordTrap(v);
      //InitMfBordTri(v);
    }

  //! Iterative merging of two fuzzy sets to define the hierarchical family 
  //! of partitions. 
  //! This function definition depends on the MF shape. It is defined for 
  //! triangular shaped ones in the file hfpintri.cpp
  void SelectMf(void); 


  public :

  //! Build a partition of 'n' MF using the vertex array. If 'Strap' is true 
  //! the partition is ended by trapezoidal fuzzy sets, else by triangular ones. 
  //! This function defenition depends on the MF shape. It is defined for 
  //! triangular shaped ones in the file hfpintri.cpp. 
  void MfInstall(int n, int Strap = false);

  //! Constructor of a 'nmf' MFs partition induced from data 't' (of size 'n') 
  //! using the type of hierarchy 'hier': 1 for hfp, 2 for kmeans, 3 for regular. 
  //! The initial partition is built using the unique values, 
  //! with tol being the equality threshold, when 'tol' > 1 it serves as the number 
  //! of groups for the kmeans algorithm.
  INHFP(double *t, int n, int hier, int nmf, double tol = TOLERANCE);

  //! Constructor used with a configuration file
  INHFP(ifstream & f, int bufsize, int num, char * s_hier, int DistSymb, int ng); 

  INHFP(int num, double min, double max) : FISIN(0)
    {
      char tmp[20];

      Init();
      active = true;
      if(NbVarG >= num  && VarNameG != NULL && VarNameG[num] != NULL)
	sprintf(tmp, "%s", VarNameG[num]);
      else sprintf(tmp, "Variable %d", num + 1);
      SetName(tmp);
      ValInf = min;
      ValSup = max;
      NmfMax = 7;
     } 

  ~INHFP() 
    {
      delete [] Val;
      delete [] Unique;
      delete [] Occur;
      if (Dist != NULL)
	{
	  for(int i = 0; i < NbUnique; i++)
	    delete [] Dist[i];
	  delete [] Dist;
	  Dist=NULL;
	}
      for(int i = 0; i < Nvertices; i++) delete [] Vertices[i];
      delete [] Vertices;
    }

  //! Generation of the hfp hierarchy of partitions. 
  void InMerge(void);  

  //! Generation of the regular grid hierarchy of partitions. 
  int HRegular(void); 

  //! Generation of the k-means hierarchy of partitions. 
  int HKmeans(void);
  
  //! Function used by FISHFP::InitSystem in complement of the INHFP constructor 
  //! from a file. 'T' is the sample data array containing 'n' items. 
  //! The initial partition is built using the unique values, 
  //! with 'thres', when 'thres'<1, or using the kmeans with 'thres' groups. 
  //! Do anything for inactive variables.
  void InitCont(double * T, int nb, double thres);

  //! Print a HFP configuration file
  virtual void PrintCfgHfp(int num, FILE *f = stdout);

  //! Print the distance array. 'num' is used to display the input number. 
  //! This is mainly a debug function.
  void PrintDist(int num); 

 
  //! Return the sum of the distances attached to the merging of the MF number 's'. 
  //! The flag before indicates that the merging is to be done 819 or has been done (0).
  double DistSum(int s, int before);  

  //! Check if the requested number of MF is compatible with the number of stored vertices, 
  //! and initialize all the MF * to NULL.
  int SetNmf(void);

  int InitNmfNvertices(void);

  //! To handle the vertex array and file
  virtual void StoreVertices(FILE * f = stdout);
  //! To handle the vertex array and file
  virtual void ReadVertices(ifstream *f, int bufsize);

  //! Return the value of 'Nvertices'
  int GetNvertices(void) const { return Nvertices; }

  //! Return the value of 'NmfMax'
  int GetNumberOfMF() const { return NmfMax; }

  //! Set the value of 'NmfMax' to 'number'
  void SetNumberOfMF( int number ) { NmfMax = number; }  

  //! Return the hierarchy type
  int GetHierType() const { return HierType; }

  const char * GetStrHierType() const
    {
      switch(HierType)
	{
	case T_HFP: return S_HFP;
	case T_KMEANS: return S_KMEANS;
	case T_REG: return S_REG;
	default : return NULL;
	}
    }
  //! Set the hierarchy type
  void SetHierType(int t) { HierType = t; }

  void SetHierType(const char * s) 
    {
      if(! strcmp(s, S_HFP)) HierType = T_HFP;
      else if (! strcmp(s, S_KMEANS)) HierType = T_KMEANS;
      else if (! strcmp(s, S_REG)) HierType = T_REG;
    }

  //! Build normalized FISIN from the centers (cent) with n MF
  //! Compute the partition coefficient (pc) and the partition entropy (pe) indices
  void ComputeIndices(double *cent, int n, double &pc, double &pe);
};
#endif
//*********************     End of HFPIN.H      ***********************
